"""
feeluown.app.config
~~~~~~~~~~~~~~~~~

App configuration items.
"""

from feeluown.config import Config


def create_config() -> Config:
    config = Config()
    config.deffield('DEBUG', type_=bool, desc='是否为调试模式')
    config.deffield('VERBOSE', default=0, type_=int, desc='日志详细程度')
    config.deffield('RPC_PORT', default=23333, type_=int, desc='RPC 端口')
    config.deffield('PUBSUB_PORT', default=23334, type_=int, desc='PUBSUB 端口')
    config.deffield('WEB_PORT', default=23332, type_=int, desc='WEB 服务端口')
    config.deffield('ENABLE_WEB_SERVER', default=False, type_=bool)
    config.deffield('MODE', default=0x0000, desc='CLI or GUI 模式')
    config.deffield('THEME', default='auto', desc='auto/light/dark')
    config.deffield('ENABLE_NEW_HOMEPAGE', default=True, type_=bool)
    config.deffield(
        'NEW_HOMEPAGE_SETTINGS',
        type_=dict,
        default={
            'contents': [
                {'name': 'RecListDailySongs', 'provider': 'netease'},
                # {'name': 'RecACollectionOfSongs', 'provider': 'qqmusic'},
                {'name': 'RecListDailyPlaylists', 'provider': 'qqmusic'},
                {'name': 'RecACollectionOfVideos', 'provider': 'bilibili'},
            ]
        },
        desc='主页配置'
    )
    config.deffield('MPV_AUDIO_DEVICE', default='auto', desc='MPV 播放设备')
    config.deffield('COLLECTIONS_DIR', desc='本地收藏所在目录')
    config.deffield(
        'FORCE_MAC_HOTKEY',
        desc='强制开启 macOS 全局快捷键功能',
        warn='Will be remove in version 3.0'
    )
    config.deffield('LOG_TO_FILE', desc='将日志输出到文件中')
    config.deffield('AUDIO_SELECT_POLICY', default='hq<>')
    config.deffield('VIDEO_SELECT_POLICY', default='hd<>')
    config.deffield('ALLOW_LAN_CONNECT', type_=bool, default=False, desc='是否可以从局域网连接服务器')
    config.deffield('PROVIDERS_STANDBY', type_=list, default=None, desc='')

    # YTDL related fields are deprecated since v4.1.9. Disable them by default.
    config.deffield(
        'ENABLE_YTDL_AS_MEDIA_PROVIDER',
        type_=bool,
        default=False,
        desc='(Deprecated) YTDL 作为备用资源'
    )
    # For example::
    #    [
    #        {
    #            'name': 'match_source',
    #            'pattern': 'ytmusic',
    #            'http_proxy': 'http://127.0.0.1:7890',
    #            'ytdl_options': {
    #                'socket_timeout': 2,
    #            },
    #        },
    #    ]
    config.deffield('YTDL_RULES', type_=list, default=None, desc='(Deprecated)')

    # TODO(cosven): maybe
    # 1. when it is set to 2, find standby from other providers first.
    # 2. when it is set to 3, play it's MV model instead of using MV's media.
    config.deffield('ENABLE_MV_AS_STANDBY', type_=int, default=1, desc='MV 作为备用资源')
    config.deffield('ENABLE_TRAY', type_=bool, default=True, desc='启用系统托盘')
    config.deffield(
        'NOTIFY_ON_TRACK_CHANGED', type_=bool, default=False, desc='切换歌曲时显示桌面通知'
    )
    config.deffield('NOTIFY_DURATION', type_=int, default=3000, desc='桌面通知保留时长(ms)')
    config.deffield('PLAYBACK_CROSSFADE', type_=bool, default=False, desc='播放暂停淡入淡出')
    config.deffield(
        'PLAYBACK_CROSSFADE_DURATION', type_=int, default=500, desc='淡入淡出持续时间'
    )
    config.deffield(
        'OPENAI_API_BASEURL',
        type_=str,
        default='',
        desc='OpenAI API base url'
    )
    config.deffield('OPENAI_API_KEY', type_=str, default='', desc='OpenAI API key')
    config.deffield('OPENAI_MODEL', type_=str, default='', desc='OpenAI model name')
    # Not sure if AI_STANDBY_MATCHER may be activated unexpectedly.
    # And it may burn some money, let user enable it manually.
    config.deffield('ENABLE_AI_STANDBY_MATCHER', type_=bool, default=False, desc='')
    config.deffield(
        'AI_RADIO_PROMPT',
        type_=str,
        default='''\
你是一个音乐推荐系统。你根据用户的歌曲列表分析用户的喜好，给用户推荐一些歌。默认推荐5首歌。

有几个注意点
1. 不要推荐与用户播放列表中一模一样的歌曲。不要推荐用户不喜欢的歌曲。不要重复推荐。
2. 你返回的内容只应该有 JSON，其它信息都不需要。也不要用 markdown 格式返回。
3. 你推荐的歌曲需要使用类似这样的 JSON 格式
    [{"title": "xxx", "artists": ["yyy", "zzz"], "description": "推荐理由"}]
''',
        desc='AI 电台功能的提示词'
    )
    return config



"""
Model v2 design principles
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

1. as much compatible as possible
2. as less magic as possible

Thinking in Model
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Q: Object or dict to store data?
A: Object is more fiendly with lint/code-auto-completion, so we have `Models`.

Q: What data/attributes should a `Model` (not) have (use *Song* as the example)?
A:
 1. identifier/title/album_name/...    -> yes: they never change
 2. url                                -> no: it changes too fast
 3. web_url                            -> no: it can be cooked with metadata
 4. album_id/artist_id                 -> yes: all providers supports these fields
 5. mv_id/comments_id/...              -> no: only some providers supports these fields

Q: One or two or many Model for one kind of object (use *Song* as the example)?
A: Obviously, we should not have too many `Model` for one Song. One `Model` is
 not so convenient. Two **seems** to be the best option. A Brief{X}Model for
 minimum and a {X}Model for some details. Extra details can be visited by
 library.{X}_{y} method.

 Q-sub1: What attributes should a Brief{X}Model have?
 A: There are several judging rules
  1. A *human* MUST be able to identify which X it is when he saw all these attributes.
  2. Each model instance must have an unique identifier to distinguish from each other.
     The provider SHOULD identify which X it is when it knows the model type and
     the indentifier.
  3. All attributes are RECOMMENDED to be string type.
  4. Less attributes as possible when the upper rules are satisfied.

 Q-sub2: What attributes should a {X}Model have?
 A: There are several judging rules
  1. Refer to some existing spec. For example, for Song model, there is already
     an ID3 tag spec. Almost all those fields defined in ID3 tag CAN be added to
     SongModel.
  2. Refer to the provider server API spec. Usually, a provider have {x}_detail API
     for a instance, and the {X}Model is RECOMMENDED to have the same attributes.
  3. Generally, a {X}Model instance can be intialized by access one or two IO operations.
     If too many operations are needed, the design CAN be bad.

 Q-sub3: Rules for adding attributes to a {X}Model?
 A:
  1. MUST keep backward compatibility, which means all new attributes should be optional
     or they have default values.
  2. Non-string type attributes are not RECOMMENDED to add.
"""

import time
from typing import List, Optional, Tuple, Any, Union

from pydantic import (
    ConfigDict, BaseModel as _BaseModel, PrivateAttr,
    model_validator, model_serializer, Field,
)

try:
    # pydantic>=2.0
    from pydantic import field_validator
    identifier_validator = field_validator('identifier', mode='before')
    pydantic_version = 2
except ImportError:
    # pydantic<2.0
    from pydantic import validator
    identifier_validator = validator('identifier', pre=True)  # type: ignore
    pydantic_version = 1

from feeluown.utils.utils import elfhash
from .base import ModelType, ModelFlags, AlbumType, MediaFlags
from .base import SearchType  # noqa
from .model_state import ModelState


TSong = Union['SongModel', 'BriefSongModel']
TAlbum = Union['AlbumModel', 'BriefAlbumModel']
TArtist = Union['ArtistModel', 'BriefArtistModel']
TVideo = Union['VideoModel', 'BriefVideoModel']
TPlaylist = Union['PlaylistModel', 'BriefPlaylistModel']
TUser = Union['UserModel', 'BriefUserModel']


def fmt_artists_names(names: List[str]) -> str:
    """Format artists names.

    >>> fmt_artists_names(['a', 'b', 'c'])
    'a, b & c'
    >>> fmt_artists_names(['a'])
    'a'
    """
    length = len(names)
    if length == 0:
        return ''
    elif length == 1:
        return names[0]
    else:
        return ' & '.join([', '.join(names[:-1]), names[-1]])


def fmt_artists(artists: List['BriefArtistModel']) -> str:
    return fmt_artists_names([artist.name for artist in artists])


def get_duration_ms(duration):
    if duration is not None:
        seconds = duration / 1000
        m, s = seconds / 60, seconds % 60
    else:
        m, s = 0, 0
    return '{:02}:{:02}'.format(int(m), int(s))


# When a model is fully supported (with v2 design), it means
# the library has implemented all features(functions) for this model.
# You can do anything with model v2 without model v1.
#
# Also, the corresponding v1 model is deprecated.
V2SupportedModelTypes = (ModelType.song, ModelType.album, ModelType.video,
                         ModelType.artist, ModelType.playlist)


class ModelMeta:
    def __init__(self, flags, model_type):
        self.flags = flags
        self.model_type = model_type

    @classmethod
    def create(cls, model_type=ModelType.dummy, is_brief=False, is_normal=False):
        flags = ModelFlags.v2
        assert not (is_brief is is_normal is True)
        if is_brief is True:
            flags |= ModelFlags.brief
        if is_normal is True:
            flags |= ModelFlags.normal
        return cls(model_type=model_type, flags=flags)


class BaseModel(_BaseModel):
    # Do not use Model.from_orm to convert v1 model to v2 model
    # since v1 model has too much magic.
    #
    # Forbidding extra fields is good for debugging. The default behavior
    # is a little implicit. If you want to store an extra attribute on model,
    # use :meth:`cache_set` explicitly.

    # For pydantic v2.
    if pydantic_version == 2:
        model_config = ConfigDict(from_attributes=False,
                                  extra='forbid',
                                  use_enum_values=True)
    else:
        # For pydantic v1.
        class Config:
            # Do not use Model.from_orm to convert v1 model to v2 model
            # since v1 model has too much magic.
            orm_mode = False
            # Forbidding extra fields is good for debugging. The default behavior
            # is a little implicit. If you want to store an extra attribute on model,
            # use :meth:`cache_set` explicitly.
            extra = 'forbid'

    _cache: dict = PrivateAttr(default_factory=dict)
    meta: Any = ModelMeta.create()

    identifier: str
    # Before, the default value of source is 'dummy', which is too implicit.
    source: str = 'dummy'
    state: ModelState = ModelState.artificial

    @identifier_validator
    def int_to_str(cls, v):
        # Old version pydantic convert int to str implicitly.
        # Many plugins(such as netease) use int as indentifier during initialization.
        # To keep backward compatibility, convert int to str here.
        if isinstance(v, int):
            return str(v)
        return v

    def cache_get(self, key) -> Tuple[Any, bool]:
        if key in self._cache:
            value, expired_at = self._cache[key]
            if expired_at is None or expired_at >= int(time.time()):
                return value, True
        return None, False

    def cache_set(self, key, value, ttl=None):
        """
        :param int ttl: the unit is seconds.
        """
        if ttl is None:
            expired_at = None
        else:
            expired_at = int(time.time()) + ttl
        self._cache[key] = (value, expired_at)

    """
    Implement __hash__ and __eq__ so that a model can be a dict key.
    From the benchmark result, the cost of __hash__ is almost equal to __eq__.
    """
    def __hash__(self):
        id_hash = elfhash(f'{self.source}{self.identifier}'.encode())
        return id_hash * 1000 + id(type(self)) % 1000

    def __eq__(self, other):
        """Implement __hash__ and __eq__ so that model can be a dict key"""
        if not isinstance(other, BaseModel):
            return False
        return all([other.source == self.source,
                    str(other.identifier) == str(self.identifier),
                    ModelType(other.meta.model_type) == self.meta.model_type])

    def __getattr__(self, attr):
        try:
            return super().__getattr__(attr)
        except AttributeError:
            if attr.endswith('_display'):
                return getattr(self, attr[:-8])
            raise

    @model_validator(mode='before')
    def _deserialize(cls, value):
        if isinstance(value, dict):
            js = value
            if 'provider' in js:
                js['source'] = js.pop('provider', None)
            js.pop('uri', None)
            js.pop('__type__', None)
            return js
        return value

    @model_serializer(mode='wrap')
    def _serialize(self, f):
        from feeluown.library import reverse

        js = f(self)
        # FIXME: pydantic >=2.11 has a bug that it may call this method
        # multiple times -> https://github.com/pydantic/pydantic/issues/11505
        # So the `js` may not have `meta` and `state` field sometimes.
        # What's even more buggy is that it does not raise exception when
        # `js` does not have `meta` and `state`.
        #
        # For example, SongModel has a field album with TAlbum type,
        # pydantic may try to serialize the model with BriefAlbumModel even
        # if the model is actually AlbumModel.
        #
        # Without this workaround, the test case `test_serialize_model`
        # will fail with pydantic >=2.11. See
        # https://github.com/feeluown/FeelUOwn/pull/922 for more details.
        js.pop('meta', None)
        js.pop('state', None)
        js['provider'] = js['source']
        js['uri'] = reverse(self)
        js['__type__'] = f'feeluown.library.{self.__class__.__name__}'
        return js


class BaseBriefModel(BaseModel):
    """
    BaseBriefModel -> model display stage
    Model -> model gotten stage
    """
    meta: Any = ModelMeta.create(is_brief=True)


class BaseNormalModel(BaseModel):
    meta: Any = ModelMeta.create(is_normal=False)
    state: ModelState = ModelState.upgraded


class BriefSongModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.song, is_brief=True)
    title: str = ''
    # TODO: maybe there should be a field `artist_names`
    # which return a list of artist name.
    artists_name: str = ''
    album_name: str = ''
    duration_ms: str = ''

    def __str__(self):
        return f'{self.source}:{self.title}•{self.artists_name}'


class BriefVideoModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.video, is_brief=True)
    title: str = ''
    artists_name: str = ''
    duration_ms: str = ''


class BriefAlbumModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.album, is_brief=True)
    name: str = ''
    artists_name: str = ''


class BriefArtistModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.artist, is_brief=True)
    name: str = ''


class BriefPlaylistModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.playlist, is_brief=True)
    name: str = ''
    creator_name: str = ''


class BriefUserModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.user, is_brief=True)
    name: str = ''


class SongModel(BriefSongModel, BaseNormalModel):
    """
    ..versionadded: 3.8.11
        The `pic_url` field.
    """
    meta: Any = ModelMeta.create(ModelType.song, is_normal=True)
    title: str
    album: Optional[TAlbum] = None
    artists: List[TArtist]
    duration: int  # milliseconds
    # A playlist can consist of multiple songs and a song can have many children.
    # The differences between playlist's songs and song' children is that
    # a child of a song usually have no album and artists. They share some
    # metadata with the parent song.
    children: List['TSong'] = []

    genre: str = ''
    date: str = ''  # For example: 2020-12-11 00:00:00, 2020-12-11T00:00:00Z
    track: str = '1/1'  # The number of the track on the album.
    disc: str = '1/1'
    # Before the field pic_url is added, user needs to fetch a AlbumModel
    # to get the image url of the song, which means that there needs another
    # IO request. However, for almost every music providers, the pic url of the song
    # can be fetched in get_song_detail API. So one IO request can be saved
    # to fetch a image url of the song.
    pic_url: str = ''
    media_flags: MediaFlags = MediaFlags.not_sure

    def model_post_init(self, _):
        super().model_post_init(_)
        self.artists_name = fmt_artists(self.artists)
        self.album_name = self.album.name if self.album else ''
        self.duration_ms = get_duration_ms(self.duration)

    def __str__(self):
        return f'{self.source}:{self.title}•{self.artists_name}'


class UserModel(BriefUserModel, BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.user, is_normal=True)
    name: str = ''
    avatar_url: str = ''


class BriefCommentModel(BaseBriefModel):
    meta: Any = ModelMeta.create(ModelType.comment, is_brief=True)
    user_name: str = ''
    content: str = ''


class CommentModel(BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.comment, is_normal=True)
    user: BriefUserModel
    content: str
    #: -1 means that the provider does not have such data
    liked_count: int
    #: unix timestamp, for example 1591695620
    time: int
    #: the parent comment which this comment replies to
    parent: Optional[BriefCommentModel] = None
    #: the root comment id
    root_comment_id: Optional[str] = None


class ArtistModel(BriefArtistModel, BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.artist, is_normal=True)
    name: str
    pic_url: str
    aliases: List[str]
    hot_songs: List[BriefSongModel]
    description: str
    song_count: int = -1
    album_count: int = -1
    mv_count: int = -1


class AlbumModel(BriefAlbumModel, BaseNormalModel):
    """
    .. versionadded:: 3.8.12
        The `song_count` field.
    """
    meta: Any = ModelMeta.create(ModelType.album, is_normal=True)
    name: str
    cover: str
    type_: AlbumType = Field(default=AlbumType.standard, validate_default=True)
    artists: List[BriefArtistModel]
    # One album usually has limited songs, and many providers' album_detail API
    # can return songs list. UPDATE(3.8.12): However, we found that albums
    # return by list_artist_album API usually has all fields except songs field.
    # To solve this problem, we add a song_count field to AlbumModel.
    #
    # The song_count field should be checked first, -1 means that the count is
    # unknown. The album may has songs or not. 0 means that the album has no songs.
    # And a positive number means that the album has exact number of songs.
    # If it is unknown, the songs field shuold be checked. If it is not empty,
    # just use it (to keep backward compatibility). If it is empty,
    # check if the provider supports SupportAlbumSongsReader protocol.
    songs: List[SongModel]
    song_count: int = -1
    description: str
    released: str = ''  # format: 2000-12-27.

    def model_post_init(self, _):
        super().model_post_init(_)
        self.artists_name = fmt_artists(self.artists)


class LyricModel(BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.lyric, is_normal=True)
    content: str
    trans_content: str = ''


class VideoModel(BriefVideoModel, BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.video, is_normal=True)
    title: str
    artists: List[BriefArtistModel]
    duration: int
    cover: str
    play_count: int = -1  # -1 means unknown
    released: str = ''  # publish date. format: 2000-12-27

    def model_post_init(self, _):
        super().model_post_init(_)
        self.artists_name = fmt_artists(self.artists)
        self.duration_ms = get_duration_ms(self.duration)


class PlaylistModel(BriefPlaylistModel, BaseNormalModel):
    meta: Any = ModelMeta.create(ModelType.playlist, is_normal=True)
    # Since modelv1 playlist does not have creator field, it is set to optional.
    creator: Optional[BriefUserModel] = None
    name: str
    cover: str
    description: str
    play_count: int = -1  # -1 means unknown
    created: str = ''  # format: 2000-12-27
    updated: str = ''  # format: 2000-12-27

    def model_post_init(self, _):
        super().model_post_init(_)
        if self.creator is not None:
            self.creator_name = self.creator.name


class SimpleSearchResult(_BaseModel):
    q: str
    songs: List[TSong] = []
    albums: List[TAlbum] = []
    artists: List[TArtist] = []
    playlists: List[TPlaylist] = []
    videos: List[TVideo] = []
    source: str = ''
    err_msg: str = ''


_type_modelcls_mapping = {
    ModelType.song: (SongModel, BriefSongModel),
    ModelType.album: (AlbumModel, BriefAlbumModel),
    ModelType.artist: (ArtistModel, BriefArtistModel),
    ModelType.video: (VideoModel, BriefVideoModel),
    ModelType.playlist: (PlaylistModel, BriefPlaylistModel),
    ModelType.user: (UserModel, BriefUserModel),
}


def get_modelcls_by_type(model_type: ModelType, brief=False):
    # return None when there is no brief model for them.
    # THINKING: maybe LyricModel and CommentModel will never has
    # a corresponding brief model.
    if model_type in _type_modelcls_mapping:
        modelcls, brief_modelcls = _type_modelcls_mapping[model_type]
        if brief:
            return brief_modelcls
        return modelcls
    return None



from typing import TYPE_CHECKING, Optional, cast

from PyQt5.QtCore import QEvent, Qt
from PyQt5.QtGui import QColor, QPainter, QResizeEvent
from PyQt5.QtWidgets import QWidget, QHBoxLayout

from feeluown.gui.helpers import esc_hide_widget

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class AppOverlayContainer(QWidget):
    """Base overlay widget that can be shown on top of other widgets.

    Features:
    - Semi-transparent background
    - Auto-resizing to parent
    - Click outside to close

    :param adhoc: If True, the overlay is destroyed when hidden.
    """

    def __init__(self, app: 'GuiApp', body: QWidget, parent: Optional[QWidget] = None,
                 adhoc=False):
        super().__init__(parent=parent)
        self._app = app
        self._adhoc = adhoc

        self.body = body
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(100, 80, 100, 80)
        self._layout.addWidget(self.body)
        self.setFocusPolicy(Qt.ClickFocus)
        # Add ClickFocus for the body so that when Overlay will not
        # get focus when user click the body.
        self.body.setFocusPolicy(Qt.ClickFocus)
        self._app.installEventFilter(self)

        esc_hide_widget(self)

    def showEvent(self, event):
        """Resize to parent when shown"""
        self.resize(self._app.size())
        super().showEvent(event)
        self.raise_()

    def hideEvent(self, event):
        """Uninstall event filter when hidden"""
        super().hideEvent(event)
        if self._adhoc is True:
            self.deleteLater()

    def paintEvent(self, event):
        """Draw semi-transparent background"""
        painter = QPainter(self)
        painter.fillRect(self.rect(), QColor(0, 0, 0, 100))

    def eventFilter(self, obj, event):
        """Handle parent resize events"""
        if self.isVisible() and obj == self._app and event.type() == QEvent.Resize:
            event = cast(QResizeEvent, event)
            self.resize(event.size())
        return super().eventFilter(obj, event)

    def focusInEvent(self, event):
        self.hide()
        super().focusInEvent(event)



from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QLabel


class BaseHeader(QLabel):
    def __init__(self, *args, font_size=13, **kwargs):
        super().__init__(*args, **kwargs)

        font = self.font()
        font.setPixelSize(font_size)
        self.setFont(font)


class LargeHeader(BaseHeader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, font_size=20, **kwargs)

        font = self.font()
        font.setWeight(QFont.DemiBold)
        self.setFont(font)


class MidHeader(BaseHeader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, font_size=16, **kwargs)



import asyncio
import logging
import os
import tempfile
import time
import uuid
from functools import partial
from hashlib import md5

from PyQt5.QtCore import QUrl
from PyQt5.QtGui import QDesktopServices, QImage

from feeluown.consts import CACHE_DIR


logger = logging.getLogger(__name__)


class ImgManager(object):
    """图片下载、缓存管理

    TOOD: 该模块相关逻辑需要重新梳理
    """
    def __init__(self, app):
        super().__init__()
        self._app = app
        self.cache = _ImgCache(self._app)

    def get_from_cache(self, img_name):
        fpath = self.cache.get(img_name)
        if fpath is not None:
            with open(fpath, 'rb') as f:
                content = f.read()
            return content
        return None

    async def get(self, img_url, img_name):
        if not img_url:
            return None
        if img_url.startswith('fuo://local'):
            # Before, `models.uri.resolve` is uesd to handle these non-std paths,
            # and it is not elegant in fact :(
            # HACK(cosven): please think about a better way in the future.
            provider = self._app.library.get('local')
            if provider is None:
                return None
            return provider.handle_with_path(img_url[11:])
        fpath = self.cache.get(img_name)
        if fpath is not None:
            with open(fpath, 'rb') as f:
                content = f.read()
            self.cache.update(img_name)
            return content
        event_loop = asyncio.get_event_loop()
        try:
            # May return None.
            res = await event_loop.run_in_executor(
                None,
                partial(self._app.request.get, img_url))
        except:  # noqa
            res = None
            logger.error(f'Download image failed, url:{img_url}')
        if res is None:
            return
        fpath = self.cache.create(img_name)
        self.save(fpath, res.content)
        return res.content

    def get_from_files(self, img_url, img_name) -> bytes:
        logger.info('extract image from {}'.format(img_url))
        if img_url.endswith('mp3') or img_url.endswith('ogg') or img_url.endswith('wma'):
            from mutagen.mp3 import EasyMP3
            metadata_mp3 = EasyMP3(img_url)
            tags_mp3 = metadata_mp3.tags
            assert tags_mp3 is not None
            content = tags_mp3._EasyID3__id3._DictProxy__dict['APIC:'].data
        elif img_url.endswith('m4a'):
            from mutagen.easymp4 import EasyMP4
            metadata_mp4 = EasyMP4(img_url)
            tags_mp4 = metadata_mp4.tags
            assert tags_mp4 is not None
            content = tags_mp4._EasyMP4Tags__mp4._DictProxy__dict['covr'][0]
        else:
            raise Exception('Unsupported file type')
        return content

    def save(self, fpath, content):
        try:
            with open(fpath, 'wb') as f:
                f.write(content)
        except Exception:
            logger.exception('save image file failed')


class _ImgCache(object):
    '''Save img in cache dir.

    Each image is saved with a hash ``name``, which contain img last used
    timestamp.
    '''
    MAX_TOTAL_NUMBER = 100

    def __init__(self, app):
        super().__init__()

        self._app = app

    def _hash(self, img_name):
        pure_url = img_name.split('?')[0]
        return md5(pure_url.encode('utf-8')).hexdigest()

    def _gen_fname(self, hname):
        ts_str = str(int(time.time()))
        return hname + '-' + ts_str

    def create(self, img_name):
        '''return img file path'''
        hname = self._hash(img_name)
        fname = self._gen_fname(hname)
        logger.debug('create img cache for %s' % img_name)
        return self._get_path(fname)

    def update(self, img_name):
        hname = self._hash(img_name)
        new_fname = self._gen_fname(hname)
        new_fpath = self._get_path(new_fname)
        old_fpath = self.get(img_name)
        os.rename(old_fpath, new_fpath)
        logger.debug('update img cache for %s' % img_name)

    def get(self, img_name):
        hname = self._hash(img_name)
        for fname in os.listdir(CACHE_DIR):
            if fname.startswith(hname):
                logger.debug('get img cache for %s' % img_name)
                return self._get_path(fname)
        return None

    def delete(self, img_name):
        fpath = self.get(img_name)
        if fpath is not None:
            return os.remove(fpath)
        return False

    def _get_path(self, fname):
        return os.path.join(CACHE_DIR, fname)


def open_image(img: QImage):
    tmpdir = tempfile.gettempdir()
    uid = str(uuid.uuid1())
    name = f'feeluown-img-{uid}.png'
    filepath = os.path.join(tmpdir, name)
    if img.save(filepath):
        QDesktopServices.openUrl(QUrl.fromLocalFile(filepath))



from feeluown.app.gui_app import GuiApp
from feeluown.collection import CollectionType, Collection
from feeluown.library import ModelType
from feeluown.utils.reader import wrap
from feeluown.gui.page_containers.table import Renderer

from feeluown.gui.base_renderer import TabBarRendererMixin


async def render(req, identifier, **kwargs):
    app: GuiApp = req.ctx['app']
    ui = app.ui
    tab_index = int(req.query.get('tab_index', 0))

    coll = app.coll_mgr.get(identifier)

    mixed = False
    model_type = None
    if coll.type is CollectionType.sys_library:
        mixed = True
    else:
        types = set()
        for model in coll.models:
            types.add(model.meta.model_type)
            if len(types) >= 2:
                mixed = True
                break
        if types:
            model_type = types.pop()
        else:
            model_type = ModelType.song

    if mixed is True:
        ui.right_panel.set_body(ui.right_panel.scrollarea)
        table_container = ui.right_panel.table_container
        renderer = LibraryRenderer(app, tab_index, coll)
        await table_container.set_renderer(renderer)
    else:
        ui.right_panel.show_collection(coll, model_type)


class LibraryRenderer(Renderer, TabBarRendererMixin):
    """
    This Renderer was used to render library collection particularlly.

    TODO: create a more elegant renderer for mixed collection.
    NOTE(cosven): I think mixed collection should be rendered in single page
    without tab.
    """
    def __init__(self, app, tab_index, coll: Collection):
        self._app = app
        self._coll = coll
        self.tab_index = tab_index
        self.tabs = self.default_tabs()

    async def render(self):
        coll = self._coll
        self.meta_widget.show()
        if coll.type is CollectionType.sys_library:
            self.meta_widget.title = '音乐库'
        else:
            self.meta_widget.title = coll.name
        self.render_tab_bar()
        self.render_models()

        def remove_song(model):
            # TODO(cosven): the whole UI is refreshed after a model is removed,
            # ideally, only part of the UI is refreshed. For example, a user scroll
            # to the bottom of the list view, when the last model is removed,
            # the UI is refreshed and the the user needs to scroll to the bottom again.
            self._coll.remove(model)
            self.render_models()
            self._app.show_msg('移除歌曲成功')

        if self.tabs[self.tab_index][1] is ModelType.song:
            self.songs_table.remove_song_func = remove_song

    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(page=f'/colls/{self._coll.identifier}',
                               query={'tab_index': tab_index})

    def render_models(self):
        _, mtype, show_handler = self.tabs[self.tab_index]
        models = [model for model in self._coll.models
                  if model.meta.model_type == mtype]
        reader = wrap(models)
        show_handler(reader)



import asyncio
import logging

from feeluown.server import handle_request
from feeluown.server.session import SessionLike
from .protocol import FuoServerProtocol, ProtocolType, PubsubProtocol


logger = logging.getLogger(__name__)


class FuoServer:
    def __init__(self, app, protocol_type):
        self._app = app
        self.protocol_type = protocol_type
        self._loop = None

    async def run(self, host, port):
        loop = asyncio.get_event_loop()
        self._loop = loop

        try:
            await loop.create_server(self.protocol_factory, host, port)
        except OSError as e:
            raise SystemExit(str(e)) from None

        name = 'unknown'
        if self.protocol_type is ProtocolType.rpc:
            name = 'PRC'
        elif self.protocol_type is ProtocolType.pubsub:
            name = 'Pub/Sub'
        logger.info('%s server run at %s:%d', name, host, port)

    def protocol_factory(self):
        if self.protocol_type is ProtocolType.rpc:
            protocol_cls = FuoServerProtocol
        else:
            protocol_cls = PubsubProtocol
        return protocol_cls(handle_req=self.handle_req,
                            loop=self._loop,)

    async def handle_req(self, req, session: SessionLike):
        return await handle_request(req, self._app, session)



from collections import deque
from typing import TYPE_CHECKING, List

from feeluown.library import SupportsSongSimilar, BriefSongModel

if TYPE_CHECKING:
    from feeluown.app import App


class Radio:
    def __init__(self, app: 'App', songs: List[BriefSongModel]):
        self._app = app
        self._stack = deque(songs)
        # B is a similar song of A. Also, A may be a similar song of B.
        # The songs_set store all songs to avoid fetching duplicate songs.
        self._songs_set = set(songs)

    def fetch_songs_func(self, number):
        """implement fm.fetch_songs_func

        TODO: We have to design a better similar-song-finding algorithm and
        we should adjust the similarity of a song by user feedback.

        Currently, we have the following ways
        1. similar songs (most providers have this functionality)
        2. a random song in the same album or similar album
        3. popular songs sang by the same artist or similar artist
        """

        valid_songs = []

        # fetch at least $number similar songs
        while len(valid_songs) < number:
            if not self._stack:
                break
            song = self._stack.popleft()
            # User can mark a song as 'dislike' by removing it from playlist.
            if song not in self._app.playlist.list():
                continue
            provider = self._app.library.get(song.source)
            # Provider is ensured to SupportsSongsimilar during creating.
            if not isinstance(provider, SupportsSongSimilar):
                continue
            songs = provider.song_list_similar(song)
            for song in songs:
                if song not in self._songs_set:
                    valid_songs.append(song)

        # TODO: sort the songs by similarity
        for song in valid_songs:
            self._stack.append(song)
            self._songs_set.add(song)
        return valid_songs


class SongRadio:
    """SongRadio recommend songs based on a song."""

    def __init__(self, app: 'App', song):
        self._app = app
        self.root_song = song
        self._radio = Radio(app, [song])

    @classmethod
    def create(cls, app, song):
        provider = app.library.get(song.source)
        if provider is not None and isinstance(provider, SupportsSongSimilar):
            return cls(app, song)
        raise ValueError('the provider must support list similar song')

    def fetch_songs_func(self, number):
        return self._radio.fetch_songs_func(number)


class SongsRadio:
    def __init__(self, app: 'App', songs: List[BriefSongModel]):
        self._app = app
        self._radio = Radio(self._app, songs)

    def fetch_songs_func(self, number):
        return self._radio.fetch_songs_func(number)



import logging
from typing import TYPE_CHECKING

from feeluown.excs import ResourceNotFound, ModelNotFound
from feeluown.library import reverse, SongModel
from feeluown.player import Metadata, MetadataFields
from feeluown.utils import aio

if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)


class MetadataAssembler:
    """Cook and fetch metadata for songs and videos"""
    def __init__(self, app: 'App'):
        self._app = app

    @staticmethod
    def cook_basic_metadata_for_song(song):
        return Metadata({
            MetadataFields.uri: reverse(song),
            MetadataFields.source: song.source,
            MetadataFields.title: song.title_display or '',
            # The song.artists_name should return a list of strings
            MetadataFields.artists: [song.artists_name_display or ''],
            MetadataFields.album: song.album_name_display or '',
        })

    async def fetch_from_song(self, song):
        empty_result = ('', '', None)
        try:
            usong: SongModel = await aio.wait_for(
                aio.run_fn(self._app.library.song_upgrade, song),
                timeout=1,
            )
        except ResourceNotFound:
            return empty_result
        except TimeoutError:  # noqa
            logger.warning(f"fetching song's meta timeout, song:'{song.title}'")
            return empty_result
        except:  # noqa
            logger.exception(f"fetching song's meta failed, song:'{song.title}'")
            return empty_result
        return usong.pic_url, usong.date, usong.album

    async def fetch_from_album(self, album):
        empty_result = ('', '')
        try:
            album = await aio.wait_for(
                aio.run_fn(self._app.library.album_upgrade, album),
                timeout=1
            )
        except ResourceNotFound:
            return empty_result
        except TimeoutError:  # noqa
            logger.warning(f"fetching album's meta timeout, album:'{album.name}'")
            return empty_result
        except:  # noqa
            logger.exception(f"fetching album meta failed, album:'{album.name}'")
            return empty_result
        return album.cover, album.released

    async def prepare_for_song(self, song):
        metadata = self.cook_basic_metadata_for_song(song)

        artwork, released, album = await self.fetch_from_song(song)
        if not (artwork and released) and album is not None:
            album_cover, album_released = await self.fetch_from_album(album)
            # Try to use album meta first.
            artwork = album_cover or artwork
            released = album_released or released
        metadata[MetadataFields.artwork] = artwork
        metadata[MetadataFields.released] = released

        return metadata

    async def prepare_for_video(self, video):
        metadata = Metadata({
            # The value of model v1 title_display may be None.
            MetadataFields.title: video.title_display or '',
            MetadataFields.source: video.source,
            MetadataFields.uri: reverse(video),
        })
        try:
            video = await aio.run_fn(self._app.library.video_upgrade, video)
        except ModelNotFound as e:
            logger.warning(f"can't get cover of video due to {str(e)}")
        else:
            metadata[MetadataFields.artwork] = video.cover
        return metadata



"""
HELP: I do not know how to design exception classes,
as a result, these interfaces can be changed frequently.
"""
from enum import Enum

from requests.exceptions import RequestException


class FuoException(Exception):
    pass


class LibraryException(FuoException):
    pass


class ProviderIOError(LibraryException, RequestException):
    """Read/write data from/to provider failed

    currently, all providers use requests to send http request,
    and many Requestexception are not catched, so ProviderIOError
    inherit RequestException.
    """

    def __init__(self, message='', provider=None):
        super().__init__(message)

        self.message = message
        self.provider = provider

    def __str__(self):
        if self.provider is None:
            return self.message
        return '[{}] {}'.format(self.provider, self.message)


class CreateReaderFailed(ProviderIOError):
    """(DEPRECATED) use ProviderIOError instead"""


class ReaderException(ProviderIOError):
    """(DEPRECATED) use ProviderIOError instead"""


class ReadFailed(ProviderIOError):
    """(DEPRECATED) use ProviderIOError instead"""


class ResourceNotFound(LibraryException):
    class Reason(Enum):
        """
        .. versionadded:: v4.0
        """
        not_found = 'not_found'
        # The provider does implement corresponding protocol.
        not_supported = 'not_supported'

    def __init__(self, *args, reason=Reason.not_found, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.reason = reason


class ProviderAlreadyRegistered(LibraryException):
    pass


class ModelNotFound(ResourceNotFound):
    """Model is not found

    For example, a model identifier is invalid.
    Maybe ResourceNotFound is enough and this exception should be removed.

    .. versionadded:: 3.7.7
    """


class MediaNotFound(ResourceNotFound):
    class Reason(Enum):
        not_found = 'not_found'
        # Song/video has children and children have medias. Song/video itself does
        # not have any media.
        check_children = 'check_children'

    def __init__(self, *args, reason=Reason.not_found, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.reason = reason


class NoUserLoggedIn(LibraryException):
    """(DEPRECATED) return None when there is no user logged in"""



import json
import logging
from typing import TYPE_CHECKING, List

from openai import AsyncOpenAI
from PyQt5.QtCore import QSize, Qt, QRectF, pyqtSignal
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import (
    QHBoxLayout, QVBoxLayout, QWidget, QLabel, QScrollArea, QPlainTextEdit,
    QFrame, QSizePolicy,
)
from PyQt5.QtGui import QPainterPath
from PyQt5.QtGui import QTextOption

from feeluown.ai import a_handle_stream
from feeluown.utils.aio import run_afn_ref
from feeluown.library import fmt_artists_names
from feeluown.library.text2song import create_dummy_brief_song
from feeluown.gui.helpers import palette_set_bg_color
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.header import MidHeader
from feeluown.gui.components.overlay import AppOverlayContainer


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


QUERY_PROMPT = '''你是一个音乐播放器助手。'''
EXTRACT_PROMPT = '''\
提取歌曲信息，歌手名为空的话，你需要补全，每首歌一行 JSON，用类似下面这样的格式返回
    {"title": "t1", "artists": ["a1", "a11"], "description": "推荐理由1"}
    {"title": "t2", "artists": ["a11"], "description": "推荐理由2"}

注意，你返回的内容只应该有几行 JSON，其它信息都不需要。也不要用 markdown 格式返回。
'''


class ChatContext:
    def __init__(self, model: str, client: AsyncOpenAI, messages: List):
        self.model = model
        self.client = client
        self.messages = messages

    async def send_message(self):
        return await self.client.chat.completions.create(
            model=self.model,
            messages=self.messages,
            stream=True,
            stream_options={'include_usage': True},
        )


class ChatInputEditor(QPlainTextEdit):
    """Custom editor for chat input with Enter key handling"""

    enter_pressed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWordWrapMode(QTextOption.WrapAtWordBoundaryOrAnywhere)
        self.setMinimumHeight(30)
        self.setMaximumHeight(300)  # TODO: set maximum height based on parent size
        self.textChanged.connect(self.adjust_height)
        self.adjust_height()
        # The size policy matters
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

    def sizeHint(self) -> QSize:  # noqa: D102
        font_metrics = self.fontMetrics()
        line_height = font_metrics.lineSpacing()
        doc = self.document()
        line_count = doc.lineCount()
        doc_height = line_count * line_height

        # Add some padding, 10
        new_height = min(max(int(doc_height) + 10, self.minimumHeight()),
                         self.maximumHeight())
        return QSize(super().sizeHint().width(), new_height)

    def adjust_height(self):
        self.updateGeometry()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return and not event.modifiers():
            self.enter_pressed.emit()
            event.accept()
        else:
            super().keyPressEvent(event)


class RoundedLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._radius = 8
        self._padding = 8
        self.setContentsMargins(self._padding, self._padding,
                                self._padding, self._padding)
        self.setWordWrap(True)

    def paintEvent(self, event):
        painter = QPainter(self)
        path = QPainterPath()
        path.addRoundedRect(QRectF(self.rect()), self._radius, self._radius)
        # Fill background
        painter.fillPath(path, self.palette().color(self.backgroundRole()))
        super().paintEvent(event)


class Body(QWidget):
    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)
        self._app = app

        # Chat history display area
        self._history_area = QScrollArea(self)
        self._history_widget = QWidget()
        self._history_area.setWidget(self._history_widget)
        self._history_layout = QVBoxLayout(self._history_widget)

        # User input area
        self._editor = ChatInputEditor(self)
        self._editor.setPlaceholderText('在这里输入你的问题...')
        self._editor.setFrameShape(QFrame.NoFrame)
        self._editor.enter_pressed.connect(
            lambda: run_afn_ref(self.exec_user_query, self._editor.toPlainText()))
        self._msg_label = QLabel(self)
        self._msg_label.setWordWrap(True)
        self._hide_btn = TextButton('关闭窗口', self)
        self._extract_and_play_btn = TextButton('提取歌曲并播放', self)
        self._extract_10_and_play_btn = TextButton('提取10首并播放', self)
        self._send_btn = TextButton('发送（回车）', self)
        self._clear_history_btn = TextButton('清空对话', self)

        self.setup_ui()
        self._hide_btn.clicked.connect(self.hide)
        self._extract_and_play_btn.clicked.connect(
            lambda: run_afn_ref(self.extract_and_play))
        self._extract_10_and_play_btn.clicked.connect(
            lambda: run_afn_ref(self.extract_10_and_play))
        self._send_btn.clicked.connect(
            lambda: run_afn_ref(self.exec_user_query, self._editor.toPlainText()))
        self._clear_history_btn.clicked.connect(self.clear_history)

        self._chat_context = None
        self.setAutoFillBackground(True)

    def setup_ui(self):
        self._history_area.setFrameShape(QFrame.NoFrame)
        self._history_area.setAutoFillBackground(True)
        self._history_layout.setContentsMargins(0, 0, 0, 0)
        self._history_layout.setSpacing(5)
        self._history_area.setWidgetResizable(True)
        # Adjust spacing between messages
        self._history_layout.setSpacing(10)

        self._msg_label.setWordWrap(True)
        self._app.installEventFilter(self)
        self._msg_label.setTextFormat(Qt.RichText)

        self._root_layout = QVBoxLayout(self)
        self._layout = QHBoxLayout()
        self._v_layout = QVBoxLayout()
        self._btn_layout = QVBoxLayout()

        self._root_layout.addWidget(MidHeader('AI 助手'))
        self._root_layout.addLayout(self._layout)
        self._layout.addStretch(0)
        self._layout.addLayout(self._v_layout)
        self._layout.setStretch(1, 1)
        self._layout.addLayout(self._btn_layout)
        self._layout.addStretch(0)
        self._root_layout.setContentsMargins(10, 10, 10, 10)
        self._root_layout.setSpacing(10)

        # Adjust layout to add chat history area
        self._v_layout.addWidget(self._history_area)
        self._v_layout.addWidget(self._msg_label)
        self._v_layout.addWidget(self._editor)
        self._btn_layout.addWidget(self._extract_and_play_btn)
        self._btn_layout.addWidget(self._extract_10_and_play_btn)
        self._btn_layout.addWidget(self._clear_history_btn)
        self._btn_layout.addWidget(self._hide_btn)
        self._btn_layout.addStretch(0)
        self._btn_layout.addWidget(self._send_btn)

    def _create_message_label(self, role, content):
        """Create message label"""
        label = RoundedLabel()
        label.setText(content)
        label.setWordWrap(True)
        label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        label.setFrameStyle(QFrame.NoFrame)

        width_factor = 0.6 if role in ('user', 'system') else 1
        label.setMaximumWidth(int(self._history_area.width() * width_factor))
        label.setAlignment(Qt.AlignLeft)

        pal = label.palette()
        if role in ('user', 'system'):
            origin_window = pal.color(pal.Window)
            palette_set_bg_color(pal, pal.color(pal.Highlight))
            pal.setColor(pal.Text, pal.color(pal.HighlightedText))
            pal.setColor(pal.Highlight, origin_window)
            label.setPalette(pal)

        return label

    def _add_message_to_history(self, role, content):
        """Add message to chat history"""
        self._chat_context.messages.append({'role': role, 'content': content})
        label = self._create_message_label(role, content)
        self._history_layout.addWidget(label)
        self._scroll_to_bottom()

    async def exec_user_query(self, query):
        if self._chat_context is None:
            self._chat_context = self.create_chat_context()

        self._add_message_to_history('user', query)
        self.set_msg('等待 AI 返回中...', level='hint')
        try:
            stream = await self._chat_context.send_message()
        except Exception as e:  # noqa
            self.set_msg(f'调用 AI 接口失败: {e}', level='err')
            logger.exception('AI request failed')
        else:
            # Create label for AI response
            ai_label = self._create_message_label('assistant', '')
            self._history_layout.addWidget(ai_label)

            content = ''
            async for chunk in stream:
                self.set_msg('AI 返回中...', level='hint')
                # 当使用 stream_options 时，最后一个 chunk 的 choices 为空
                if chunk.choices:
                    delta_content = chunk.choices[0].delta.content or ''
                    content += delta_content
                    # Update AI response in real-time
                    ai_label.setText(content)
                    self._scroll_to_bottom()

            # Update chat context and show token usage
            assistant_message = {"role": "assistant", "content": content}
            self._chat_context.messages.append(assistant_message)
            self.show_tokens_usage(chunk)
            # Clear input box
            self._editor.clear()

    def show_tokens_usage(self, chunk):
        if not chunk:
            self.set_msg('AI 内容返回结束', level='hint')
            return

        in_tokens = chunk.usage.prompt_tokens if chunk.usage else 0
        out_tokens = chunk.usage.completion_tokens if chunk.usage else 0
        total_tokens = chunk.usage.total_tokens if chunk.usage else 0
        token_msg = f"Tokens: 输入 {in_tokens}, 输出 {out_tokens}, 合计 {total_tokens}"
        self.set_msg(f'AI 内容返回结束 ({token_msg})', level='hint')

    def set_msg(self, text, level='hint'):
        if level == 'hint':
            color = 'green'
        elif level == 'warn':
            color = 'yellow'
        else:  # err
            color = 'magenta'
        self._msg_label.setText(f'<span style="color: {color}">{text}</span>')

    def create_chat_context(self):
        return ChatContext(
            model=self._app.config.OPENAI_MODEL,
            client=self._app.ai.get_async_client(),
            messages=[{'role': 'system', 'content': QUERY_PROMPT}]
        )

    async def extract_and_play(self):
        await self._extract_and_play(EXTRACT_PROMPT)

    async def extract_10_and_play(self):
        await self._extract_and_play(f'{EXTRACT_PROMPT}\n随机提取最多10首即可')

    async def _extract_and_play(self, extract_prompt):
        """Main entry point for extracting and playing songs"""
        self._prepare_extract_context(extract_prompt)
        self.set_msg('正在让 AI 解析歌曲信息，这可能会花费一些时间...')
        try:
            stream = await self._chat_context.send_message()
            await self._process_extract_stream(stream)
        except Exception as e:
            self.set_msg(f'调用 AI 接口失败: {e}', level='err')
            logger.exception('AI request failed')

    def _prepare_extract_context(self, extract_prompt):
        """Prepare chat context for song extraction"""
        if self._chat_context is None:
            self.set_msg('没有对话上下文', level='err')
        else:
            self._add_message_to_history('user', extract_prompt)

    async def _process_extract_stream(self, stream):
        """Process the stream of extracted songs"""
        rr, rw, wtask = await a_handle_stream(stream)
        ok_count = 0
        fail_count = 0

        # 创建AI回复的标签
        ai_label = self._create_message_label('assistant', '')
        self._history_layout.addWidget(ai_label)
        content = ''
        try:
            while True:
                try:
                    line = await rr.readline()
                    line = line.decode('utf-8')
                    content += f'{line}\n'  # add newline
                    ai_label.setText(content)
                    self._scroll_to_bottom()
                    logger.debug(f'read a line: {line}')
                    if not line:
                        self.set_msg(f'解析结束，成功解析{ok_count}首歌曲，失败{fail_count}首歌。',
                                     level='hint')
                        break

                    try:
                        jline = json.loads(line)
                        title, artists = jline['title'], jline['artists']
                        artists_name = fmt_artists_names(artists)
                        song = create_dummy_brief_song(title, artists_name)
                    except Exception:
                        fail_count += 1
                        logger.exception(f'failed to parse a line: {line}')
                        self.set_msg(f'成功解析{ok_count}首歌曲，失败{fail_count}首歌',
                                     level='yellow')
                        continue

                    ok_count += 1
                    self.set_msg(f'成功解析{ok_count}首歌曲，失败{fail_count}首歌',
                                 level='hint')
                    self._app.playlist.add(song)
                    if ok_count == 1:
                        self._app.playlist.play_model(song)
                except Exception:
                    logger.exception('Error processing song')
                    break
        finally:
            assistant_message = {"role": "assistant", "content": content}
            self._chat_context.messages.append(assistant_message)
            chunk = await wtask
            self.show_tokens_usage(chunk)
            self._scroll_to_bottom()
            rw.close()
            await rw.wait_closed()
            self._editor.clear()

    def _scroll_to_bottom(self) -> None:
        """Scroll chat history to bottom."""
        self._history_area.verticalScrollBar().setValue(
            self._history_area.verticalScrollBar().maximum()
        )

    def clear_history(self):
        """Clear chat history"""
        while self._history_layout.count():
            item = self._history_layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()
        self.set_msg('')
        self._chat_context = None

    def hide(self):
        self.clear_history()
        self.parent().hide()


def create_aichat_overlay(app: 'GuiApp', parent=None) -> AppOverlayContainer:
    """Create an overlay for the AI chat"""
    body = Body(app)
    overlay = AppOverlayContainer(app, body, parent=parent)
    return overlay


if __name__ == '__main__':
    import os
    from PyQt5.QtWidgets import QWidget
    from feeluown.gui.debug import simple_layout, mock_app

    with simple_layout(theme='dark') as layout, mock_app() as app:
        app.size.return_value = QSize(600, 400)
        app.config.OPENAI_API_KEY = os.environ.get('DEEPSEEK_API_KEY')
        app.config.OPENAI_API_BASEURL = 'https://api.deepseek.com'
        app.config.OPENAI_MODEL = 'deepseek-chat'
        widget = create_aichat_overlay(app)
        widget.resize(600, 400)
        layout.addWidget(widget)
        widget.show()
        widget.body.set_msg('error', level='err')

        widget.body._chat_context = widget.body.create_chat_context()
        widget.body._add_message_to_history('user', '哈哈哈'*10)
        widget.body._add_message_to_history('xxx', '哈哈哈'*100)



import logging
from typing import TYPE_CHECKING, Optional, TypeVar, Generic

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel

from feeluown.library import (
    SupportsRecListDailyPlaylists, SupportsRecACollectionOfSongs, Collection,
    SupportsRecListDailySongs, Provider, SupportsRecACollectionOfVideos,
)
from feeluown.utils.reader import create_reader
from feeluown.utils.aio import run_fn, gather, run_afn
from feeluown.gui.widgets.header import LargeHeader
from feeluown.gui.widgets.img_card_list import (
    PlaylistCardListView,
    PlaylistCardListModel,
    PlaylistFilterProxyModel,
    PlaylistCardListDelegate,
    VideoCardListView,
    VideoCardListModel,
    VideoCardListDelegate,
)
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListView,
    SongMiniCardListDelegate,
    SongMiniCardListModel,
)
from feeluown.gui.widgets.selfpaint_btn import PlayButton
from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.gui.helpers import fetch_cover_wrapper, BgTransparentMixin

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


async def render(req, **kwargs):
    app: 'GuiApp' = req.ctx['app']

    view = View(app)
    scroll_area = ScrollArea()
    scroll_area.setWidget(view)
    app.ui.right_panel.set_body(scroll_area)
    await view.render()


class Panel(QWidget):
    _id_pixmap_cache = {}

    def __init__(self, title, body, pixmap):
        super().__init__(parent=None)

        self.icon_label = QLabel()
        self.header = LargeHeader(title)
        self.body = body

        self.icon_label.setFixedSize(20, 20)
        self.icon_label.setPixmap(pixmap)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(10)
        self._h_layout = QHBoxLayout()
        self._h_layout.setSpacing(5)
        self._layout.addLayout(self._h_layout)
        self._h_layout.addWidget(self.icon_label)
        self._h_layout.addWidget(self.header)
        self._layout.addWidget(self.body)

    @classmethod
    def get_provider_pixmap(cls, app: 'GuiApp', provider_id):
        if provider_id in cls._id_pixmap_cache:
            return cls._id_pixmap_cache[provider_id]
        pvd_ui = app.pvd_ui_mgr.get(provider_id)
        if pvd_ui is None:
            return QPixmap()
        svg = pvd_ui.get_colorful_svg()
        return QPixmap(svg).scaledToWidth(20, Qt.SmoothTransformation)

    async def render(self):
        pass


class HBody(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)


class RecPlaylistsPanel(Panel):

    def __init__(self, app: 'GuiApp', provider: SupportsRecListDailyPlaylists):
        self._provider = provider
        self._app = app
        title = '推荐歌单'
        self.playlist_list_view = PlaylistCardListView(no_scroll_v=True)
        pixmap = Panel.get_provider_pixmap(app, provider.identifier)
        super().__init__(title, self.playlist_list_view, pixmap)

    async def render(self):
        playlists = await run_fn(self._provider.rec_list_daily_playlists)
        print(playlists)
        if not playlists:
            return
        playlist_list_view = self.playlist_list_view
        playlist_list_view.show_playlist_needed.connect(
            lambda model: self._app.browser.goto(model=model)
        )
        playlist_list_view.setItemDelegate(
            PlaylistCardListDelegate(
                playlist_list_view,
                card_min_width=150,
            )
        )
        model = PlaylistCardListModel(
            create_reader(playlists), fetch_cover_wrapper(self._app),
            {p.identifier: p.name
             for p in self._app.library.list()}
        )
        filter_model = PlaylistFilterProxyModel()
        filter_model.setSourceModel(model)
        playlist_list_view.setModel(filter_model)


P = TypeVar('P', bound=Provider)


class SongsBasePanel(Panel, Generic[P]):
    """
    Base panel class for show a list of songs.
    """

    def __init__(self, title: str, app: 'GuiApp', provider: P):
        self._app = app
        self._provider = provider
        self.songs_list_view = songs_list_view = SongMiniCardListView(no_scroll_v=True)
        songs_list_view.setItemDelegate(SongMiniCardListDelegate(songs_list_view, ))
        pixmap = Panel.get_provider_pixmap(app, provider.identifier)
        super().__init__(title, songs_list_view, pixmap)

        self.play_all_btn = PlayButton()
        self._h_layout.addWidget(self.play_all_btn)
        self._h_layout.addStretch(0)

        self.play_all_btn.clicked.connect(lambda: run_afn(self._play_all))
        songs_list_view.play_song_needed.connect(self._app.playlist.play_model)

    async def _play_all(self):
        songs = await run_fn(self.songs_list_view.model().get_reader().readall)
        self._app.playlist.set_models(songs, next_=True)
        self._app.player.resume()

    def set_reader(self, reader):
        songs_model = SongMiniCardListModel.create(reader, self._app)
        self.songs_list_view.setModel(songs_model)


class RecDailySongsPanel(SongsBasePanel[SupportsRecListDailySongs]):

    def __init__(self, app: 'GuiApp', provider: SupportsRecListDailySongs):
        super().__init__('每日推荐', app, provider)

    async def render(self):
        songs = await run_fn(self._provider.rec_list_daily_songs)
        self.set_reader(songs)


class RecSongsPanel(SongsBasePanel[SupportsRecACollectionOfSongs]):

    def __init__(self, app: 'GuiApp', provider: SupportsRecACollectionOfSongs):
        super().__init__('随便听听', app, provider)

    async def render(self):
        coll: Collection = await run_fn(self._provider.rec_a_collection_of_songs)
        self.set_reader(coll.models)
        self.header.setText(coll.name)


class RecVideosPanel(Panel):
    def __init__(self, app: 'GuiApp', provider: SupportsRecACollectionOfVideos):
        self._app = app
        self._provider = provider
        self.video_list_view = video_list_view = VideoCardListView()
        video_list_view.setItemDelegate(VideoCardListDelegate(
            video_list_view,
            card_min_width=200,
        ))
        pixmap = Panel.get_provider_pixmap(app, provider.identifier)
        super().__init__('瞅瞅', video_list_view, pixmap)

        video_list_view.play_video_needed.connect(self._app.playlist.play_model)

    async def render(self):
        coll = await run_fn(self._provider.rec_a_collection_of_videos)
        videos = coll.models
        if videos:
            # TODO: maybe show all videos
            model = VideoCardListModel.create(videos[:8], self._app)
            self.video_list_view.setModel(model)
            self.header.setText(coll.name)
        else:
            self.header.setText('暂无推荐视频')
            self.video_list_view.hide()


class View(QWidget, BgTransparentMixin):

    def __init__(self, app: 'GuiApp'):
        super().__init__(parent=None)
        self._app = app

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(20, 10, 20, 0)
        self._layout.setSpacing(10)

    async def render(self):
        panels = []
        settings = self._app.config.NEW_HOMEPAGE_SETTINGS
        for content in settings.get('contents', []):
            name = content['name']
            if name == 'RecListDailySongs':
                panel = self._handle_rec_list_daily_songs(content)
                if panel is not None:
                    panels.append(panel)
            elif name == 'RecListDailyPlaylists':
                panel = self._handle_rec_list_daily_playlists(content)
                if panel is not None:
                    panels.append(panel)
            elif name == 'RecACollectionOfSongs':
                panel = self._handle_rec_a_collection_of_songs(content)
                if panel is not None:
                    panels.append(panel)
            elif name == 'RecACollectionOfVideos':
                panel = self._handle_rec_a_collection_of_videos(content)
                if panel is not None:
                    panels.append(panel)
        for panel in panels:
            self._layout.addWidget(panel)
        gather(*[panel.render() for panel in panels])

    def _handle_rec_a_collection_of_videos(self, content: dict) -> Optional[Panel]:
        source = content['provider']
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecACollectionOfVideos):
            return RecVideosPanel(self._app, provider)
        logger.warning(f'Invalid homepage content: {content}, '
                       f'provider {source} not found or not supported')
        return None

    def _handle_rec_list_daily_songs(self, content: dict) -> Optional[Panel]:
        source = content['provider']
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecListDailySongs):
            return RecDailySongsPanel(self._app, provider)
        logger.warning(f'Invalid homepage content: {content}, '
                       f'provider {source} not found or not supported')
        return None

    def _handle_rec_list_daily_playlists(self, content: dict) -> Optional[Panel]:
        source = content['provider']
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecListDailyPlaylists):
            return RecPlaylistsPanel(self._app, provider)
        logger.warning(f'Invalid homepage content: {content}, '
                       f'provider {source} not found or not supported')
        return None

    def _handle_rec_a_collection_of_songs(self, content: dict) -> Optional[Panel]:
        source = content['provider']
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecACollectionOfSongs):
            return RecSongsPanel(self._app, provider)
        logger.warning(f'Invalid homepage content: {content}, '
                       f'provider {source} not found or not supported')
        return None



# pylint: disable=import-error
import logging
import os

import dbus
import dbus.service

from feeluown.app import App
from feeluown.player import State, PlaylistShuffleMode, PlaylistRepeatMode


SupportedMimeTypes = ['audio/aac',
                      'audio/m4a',
                      'audio/mp3',
                      'audio/wav',
                      'audio/wma',
                      'audio/x-ape',
                      'audio/x-flac',
                      'audio/x-ogg',
                      'audio/x-oggflac',
                      'audio/x-vorbis']
BusName = 'org.mpris.MediaPlayer2.feeluown'
ObjectPath = '/org/mpris/MediaPlayer2'
AppInterface = 'org.mpris.MediaPlayer2'
PlayerInterface = 'org.mpris.MediaPlayer2.Player'
AppProperties = dbus.Dictionary({
    'DesktopEntry': 'FeelUOwn',
    'Identity': 'feeluown',
    'CanQuit': True,
    'CanRaise': True,
    'HasTrackList': False,
    'SupportedUriSchemes': ['http', 'file', 'fuo'],
    'SupportedMimeTypes': SupportedMimeTypes,
}, signature='sv')
RepeatModeLoopStatusMapping = {
    PlaylistRepeatMode.all: 'Playlist',
    PlaylistRepeatMode.one: 'Track',
    PlaylistRepeatMode.none: 'None',
}
LoopStatusRepeatModeMapping = {v: k for k, v in RepeatModeLoopStatusMapping.items()}


logger = logging.getLogger(__name__)


def to_dbus_position(p):
    return dbus.Int64(p * 1000 * 1000)


def to_fuo_position(p):
    return p / 1000 / 1000


def to_dbus_volume(volume):
    return round(volume / 100, 1)


def to_fuo_volume(volume):
    return int(volume * 100)


def to_dbus_playback_status(state):
    if state == State.stopped:
        status = 'Stopped'
    elif state == State.paused:
        status = 'Paused'
    else:
        status = 'Playing'
    return status


def to_dbus_metadata(metadata):
    if metadata:
        artists = metadata.get('artists', ['Unknown'])[:1]
    else:
        # make xesam:artist a one-element list to compat with KDE
        # KDE will not update artist field if the length>=1
        artists = ['']
    uri = metadata.get('uri', 'fuo://unknown/unknown/unknown')
    return dbus.Dictionary({
        # If there is no artist, we give a empty string in case mpris complains
        # 'ValueError: Unable to guess signature from an empty list'
        'xesam:artist': artists or [''],
        'xesam:url': uri,
        # trackid must be provided to make 'Seek' work.
        'mpris:trackid': f'/org/feeluown/FeelUOwn/{uri[6:]}',
        'mpris:artUrl': metadata.get('artwork', ''),
        'xesam:album': metadata.get('album', ''),
        'xesam:title': metadata.get('title', ''),
    }, signature='sv')


class Mpris2Service(dbus.service.Object):
    # pylint: disable=too-many-public-methods

    def __init__(self, app: App, bus):
        super().__init__(bus, ObjectPath)
        self._app = app

    def enable(self):
        # 根据 mpris2 规范, position 变化时，不需要发送 PropertiesChanged 信号。
        # audacious 的一个 issue 也证明了这个观点:
        #   https://redmine.audacious-media-player.org/issues/849
        #   这个 issue 主要是说频繁发送 PropertiesChanged 信号会导致
        #   GNOME 桌面消耗大量 CPU。
        self._app.player.seeked.connect(self.update_position)
        self._app.player.duration_changed.connect(self.update_duration)
        self._app.player.state_changed.connect(self.update_playback_status)
        self._app.playlist.playback_mode_changed.connect(self.update_playback_mode)
        self._app.player.metadata_changed.connect(self.update_song_props)

    def disable(self):
        pass

    def update_playback_status(self, state):
        status = to_dbus_playback_status(state)
        self.PropertiesChanged(PlayerInterface,
                               {'PlaybackStatus': status}, [])

    def update_playback_mode(self, _):
        props = {
            'LoopStatus': RepeatModeLoopStatusMapping[self._app.playlist.repeat_mode],
            'Shuffle': self._app.playlist.shuffle_mode is not PlaylistShuffleMode.off,
        }
        self.PropertiesChanged(PlayerInterface, props, [])

    def update_position(self, position):
        self.Seeked(to_dbus_position(position))

    def update_duration(self, duration):
        if duration <= 0:  # Duration can be 0 when media is changed.
            return
        length = to_dbus_position(duration or 0)
        metadata = to_dbus_metadata(self._app.player.current_metadata)
        metadata['mpris:length'] = length
        props = dbus.Dictionary({'Metadata': metadata})
        self.PropertiesChanged(PlayerInterface, props, [])

    def update_song_props(self, metadata):
        props = dbus.Dictionary({'Metadata': to_dbus_metadata(metadata)})
        self.PropertiesChanged(PlayerInterface, props, [])

    def get_player_properties(self):
        return dbus.Dictionary({
            'Metadata': to_dbus_metadata(self._app.player.current_metadata),
            'Rate': 1.0,
            'MinimumRate': 1.0,
            'MaximumRate': 1.0,
            'CanGoNext': True,
            'CanGoPrevious': True,
            'CanControl': True,
            'CanSeek': True,
            'CanPause': True,
            'CanPlay': True,
            'Position': to_dbus_position(self._app.player.position or 0),
            'LoopStatus': RepeatModeLoopStatusMapping[self._app.playlist.repeat_mode],
            'Shuffle': self._app.playlist.shuffle_mode is not PlaylistShuffleMode.off,
            'PlaybackStatus': to_dbus_playback_status(self._app.player.state),
            'Volume': to_dbus_volume(self._app.player.volume),
        }, signature='sv', variant_level=2)

    # ##########################
    # implement mpris2 interface
    # ##########################
    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def Play(self):
        self._app.player.resume()

    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def Pause(self):
        self._app.player.pause()

    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def Next(self):
        self._app.playlist.next()

    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def Previous(self):
        self._app.playlist.previous()

    @dbus.service.method(PlayerInterface, in_signature='s', out_signature='')
    def OpenUri(self, Uri):
        pass

    @dbus.service.method(PlayerInterface, in_signature='ox', out_signature='')
    def SetPosition(self, TrackId, Position):
        self._app.player.position = to_fuo_position(Position)

    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def PlayPause(self):
        self._app.player.toggle()

    @dbus.service.method(PlayerInterface, in_signature='x', out_signature='')
    def Seek(self, Offset):
        self._app.player.position = to_fuo_position(Offset)

    @dbus.service.method(PlayerInterface, in_signature='', out_signature='')
    def Stop(self):
        self._app.player.stop()

    @dbus.service.signal(PlayerInterface, signature='x')
    def Seeked(self, Position):
        pass

    @dbus.service.signal(dbus.PROPERTIES_IFACE, signature='sa{sv}as')
    def PropertiesChanged(self, interface, changed_properties,
                          invalidated_properties=None):
        pass

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='ss', out_signature='v')
    def Get(self, interface, prop):
        return self.GetAll(interface)[prop]

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='ssv')
    def Set(self, interface, prop, value):
        if prop == 'Volume':
            self._app.player.volume = to_fuo_volume(value)
        elif prop == 'LoopStatus':
            self._app.playlist.repeat_mode = LoopStatusRepeatModeMapping[value]
        elif prop == 'Shuffle':
            shuffle_mode = PlaylistShuffleMode.songs if value else \
                PlaylistShuffleMode.off
            self._app.playlist.shuffle_mode = shuffle_mode
        else:
            logger.info("mpris wants to set %s to %s", prop, value)

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface == PlayerInterface:
            props = self.get_player_properties()
            return props
        if interface == AppInterface:
            return AppProperties
        raise dbus.exceptions.DBusException(
            'com.example.UnknownInterface',
            f'The Foo object does not implement the {interface} interface'
        )

    @dbus.service.method(AppInterface, in_signature='', out_signature='')
    def Quit(self):
        self._app.exit()

    @dbus.service.method(AppInterface, in_signature='', out_signature='')
    def Raise(self):
        if self._app.has_gui:
            self._app.raise_()

    @dbus.service.method(dbus.INTROSPECTABLE_IFACE, in_signature='', out_signature='s')
    def Introspect(self):  # pylint: disable=arguments-differ
        current_dir_name = os.path.dirname(os.path.realpath(__file__))
        xml = os.path.join(current_dir_name, 'introspect.xml')
        with open(xml, 'r', encoding='utf-8') as f:
            contents = f.read()
        return contents



from typing import runtime_checkable, Protocol, List, Tuple, Optional, Dict
from abc import abstractmethod
from feeluown.media import Quality, Media
from feeluown.excs import NoUserLoggedIn
from feeluown.utils.dispatch import Signal
from .models import (
    BriefCommentModel, SongModel, VideoModel, AlbumModel, ArtistModel,
    PlaylistModel, UserModel, ModelType, BriefArtistModel, BriefSongModel,
    LyricModel, BriefVideoModel, BriefPlaylistModel,
)
from .collection import Collection

from .flags import Flags as PF


ID = str
_FlagProtocolMapping: Dict[Tuple[ModelType, PF], type] = {}


def check_flag(provider, model_type: ModelType, flag: PF) -> bool:
    """Check if provider supports X"""

    # A provider should declare explicitly whether it uses model v2 or not.
    if flag is PF.model_v2:
        try:
            use_model_v2 = provider.use_model_v2(model_type)
        except AttributeError:  # noqa
            # The provider may not implement the `use_model_v2` interface.
            return False
        return use_model_v2

    protocol_cls = _FlagProtocolMapping[(model_type, flag)]
    return isinstance(provider, protocol_cls)


def eq(model_type: ModelType, flag: PF):
    """Decorate a protocol class and associate it with a provider flag"""
    def wrapper(cls):
        _FlagProtocolMapping[(model_type, flag)] = cls
        return cls
    return wrapper


#
# Protocols for Song related functions.
#

@eq(ModelType.song, PF.get)
@runtime_checkable
class SupportsSongGet(Protocol):
    @abstractmethod
    def song_get(self, identifier: ID) -> SongModel:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@eq(ModelType.song, PF.similar)
@runtime_checkable
class SupportsSongSimilar(Protocol):
    @abstractmethod
    def song_list_similar(self, song: BriefSongModel) -> List[BriefSongModel]:
        """List similar songs
        """
        raise NotImplementedError


@eq(ModelType.song, PF.multi_quality)
@runtime_checkable
class SupportsSongMultiQuality(Protocol):
    @abstractmethod
    def song_list_quality(self, song: BriefSongModel) -> List[Quality.Audio]:
        """List all possible qualities

        Please ensure all the qualities are valid. `song_get_media(song, quality)`
        must not return None with a valid quality.
        """
        raise NotImplementedError

    @abstractmethod
    def song_select_media(
        self, song: BriefSongModel, policy=None
    ) -> Tuple[Media, Quality.Audio]:
        """Select a media by the quality sorting policy

        If the song has some valid medias, this method can always return one of them.
        """
        raise NotImplementedError

    @abstractmethod
    def song_get_media(
        self, song: BriefVideoModel, quality: Quality.Audio
    ) -> Optional[Media]:
        """Get song's media by a specified quality

        :return: when quality is invalid, return None
        """
        raise NotImplementedError


@eq(ModelType.song, PF.hot_comments)
@runtime_checkable
class SupportsSongHotComments(Protocol):
    def song_list_hot_comments(self, song: BriefSongModel) -> List[BriefCommentModel]:
        raise NotImplementedError


@eq(ModelType.song, PF.web_url)
@runtime_checkable
class SupportsSongWebUrl(Protocol):
    def song_get_web_url(self, song: BriefSongModel) -> str:
        raise NotImplementedError


@eq(ModelType.song, PF.lyric)
@runtime_checkable
class SupportsSongLyric(Protocol):
    def song_get_lyric(self, song: BriefSongModel) -> Optional[LyricModel]:
        """Get music video of the song
        """
        raise NotImplementedError


@eq(ModelType.song, PF.mv)
@runtime_checkable
class SupportsSongMV(Protocol):
    def song_get_mv(self, song: BriefSongModel) -> Optional[VideoModel]:
        """Get music video of the song

        """
        raise NotImplementedError


#
# Protocols for Album related functions.
#

@eq(ModelType.album, PF.get)
@runtime_checkable
class SupportsAlbumGet(Protocol):
    @abstractmethod
    def album_get(self, identifier: ID) -> AlbumModel:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@eq(ModelType.album, PF.songs_rd)
@runtime_checkable
class SupportsAlbumSongsReader(Protocol):
    @abstractmethod
    def album_create_songs_rd(self, album) -> List[SongModel]:
        raise NotImplementedError


#
# Protocols for Album related functions.
#

@eq(ModelType.artist, PF.get)
@runtime_checkable
class SupportsArtistGet(Protocol):
    @abstractmethod
    def artist_get(self, identifier: ID) -> ArtistModel:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@eq(ModelType.artist, PF.songs_rd)
@runtime_checkable
class SupportsArtistSongsReader(Protocol):
    @abstractmethod
    def artist_create_songs_rd(self, artist: BriefArtistModel):
        """Create songs reader of the artist
        """
        raise NotImplementedError


@eq(ModelType.artist, PF.albums_rd)
@runtime_checkable
class SupportsArtistAlbumsReader(Protocol):
    @abstractmethod
    def artist_create_albums_rd(self, artist: BriefArtistModel):
        """Create albums reader of the artist
        """
        raise NotImplementedError


@runtime_checkable
class SupportsArtistContributedAlbumsReader(Protocol):
    @abstractmethod
    def artist_create_contributed_albums_rd(self, artist: BriefArtistModel):
        """Create contributed albums reader of the artist
        """
        raise NotImplementedError


#
# Protocols for Video related functions.
#

@eq(ModelType.video, PF.get)
@runtime_checkable
class SupportsVideoGet(Protocol):
    @abstractmethod
    def video_get(self, identifier: ID) -> VideoModel:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@runtime_checkable
class SupportsVideoWebUrl(Protocol):
    def video_get_web_url(self, video: BriefVideoModel) -> str:
        raise NotImplementedError


@eq(ModelType.video, PF.multi_quality)
@runtime_checkable
class SupportsVideoMultiQuality(Protocol):
    @abstractmethod
    def video_list_quality(self, video: BriefVideoModel) -> List[Quality.Video]:
        raise NotImplementedError

    @abstractmethod
    def video_select_media(
            self, video: BriefVideoModel, policy=None) -> Tuple[Media, Quality.Video]:
        raise NotImplementedError

    @abstractmethod
    def video_get_media(self, video: BriefVideoModel, quality) -> Optional[Media]:
        raise NotImplementedError


#
# Protocols for Album related functions.
#

@eq(ModelType.playlist, PF.get)
@runtime_checkable
class SupportsPlaylistGet(Protocol):
    @abstractmethod
    def playlist_get(self, identifier: ID) -> PlaylistModel:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@runtime_checkable
class SupportsPlaylistCreateByName(Protocol):
    @abstractmethod
    def playlist_create_by_name(self, name) -> PlaylistModel:
        """Create playlist for user logged in.

        :raises NoUserLoggedIn:
        :raises ProviderIOError:
        """


@runtime_checkable
class SupportsPlaylistDelete(Protocol):
    @abstractmethod
    def playlist_delete(self, identifier: ID) -> bool:
        """
        :raises ModelNotFound: model not found by the identifier
        :raises ProviderIOError:
        """
        raise NotImplementedError


@eq(ModelType.playlist, PF.songs_rd)
@runtime_checkable
class SupportsPlaylistSongsReader(Protocol):
    @abstractmethod
    def playlist_create_songs_rd(self, playlist):
        raise NotImplementedError


@eq(ModelType.playlist, PF.add_song)
@runtime_checkable
class SupportsPlaylistAddSong(Protocol):
    @abstractmethod
    def playlist_add_song(self, playlist, song) -> bool:
        raise NotImplementedError


@eq(ModelType.playlist, PF.remove_song)
@runtime_checkable
class SupportsPlaylistRemoveSong(Protocol):
    @abstractmethod
    def playlist_remove_song(self, playlist, song) -> bool:
        raise NotImplementedError


#
# Protocols for None related functions.
#
@eq(ModelType.none, PF.current_user)
@runtime_checkable
class SupportsCurrentUser(Protocol):
    @abstractmethod
    def has_current_user(self) -> bool:
        """Check if there is a logged in user."""

    @abstractmethod
    def get_current_user(self) -> UserModel:
        """Get current logged in user

        :raises NoUserLoggedIn: there is no logged in user.

        .. versionchanged: 4.0
            It can return Optional[UserModel], NoUserLoggedIn makes no sense.
        """

    def get_current_user_or_none(self) -> Optional[UserModel]:
        """
        .. versionadded: 4.0
        """
        try:
            return self.get_current_user()
        except NoUserLoggedIn:
            return None


@runtime_checkable
class SupportsUserAutoLogin(Protocol):
    """Protocol for providers that support automatic login using cached credentials."""

    @abstractmethod
    def auto_login(self) -> bool:
        """Try to automatically login using cached credentials.

        Returns:
            bool: True if auto login succeeded, False otherwise
        """


@runtime_checkable
class SupportsCurrentUserChanged(Protocol):
    @property
    @abstractmethod
    def current_user_changed(self) -> Signal:
        """

        :return: Signal(UserModel)
        """
        ...


@runtime_checkable
class SupportsCurrentUserListPlaylists(Protocol):
    @abstractmethod
    def current_user_list_playlists(self):
        """
        : raises NoUserLoggedIn:
        """


#
# Protocols for current user favorites/collections
#
@runtime_checkable
class SupportsCurrentUserFavSongsReader(Protocol):
    @abstractmethod
    def current_user_fav_create_songs_rd(self):
        """
        : raises NoUserLoggedIn:
        """


@runtime_checkable
class SupportsCurrentUserFavAlbumsReader(Protocol):
    @abstractmethod
    def current_user_fav_create_albums_rd(self):
        pass


@runtime_checkable
class SupportsCurrentUserFavArtistsReader(Protocol):
    @abstractmethod
    def current_user_fav_create_artists_rd(self):
        pass


@runtime_checkable
class SupportsCurrentUserFavPlaylistsReader(Protocol):
    @abstractmethod
    def current_user_fav_create_playlists_rd(self):
        pass


@runtime_checkable
class SupportsCurrentUserFavVideosReader(Protocol):
    @abstractmethod
    def current_user_fav_create_videos_rd(self):
        pass


@runtime_checkable
class SupportsCurrentUserDislikeSongsReader(Protocol):
    """Support reading the song dislike list."""
    @abstractmethod
    def current_user_dislike_create_songs_rd(self) -> List[BriefSongModel]:
        pass


@runtime_checkable
class SupportsCurrentUserDislikeAddSong(Protocol):
    """Support adding a song to the song dislike list."""
    @abstractmethod
    def current_user_dislike_add_song(self, song: BriefSongModel) -> bool:
        """
        :return: True if the song is added to the dislike list, False otherwise.
        """


@runtime_checkable
class SupportsCurrentUserDislikeRemoveSong(Protocol):
    """Support removing a song from the song dislike list."""
    @abstractmethod
    def current_user_dislike_remove_song(self, song: BriefSongModel) -> bool:
        """
        :return: True if the song is removed from the dislike list, False otherwise.
        """


@runtime_checkable
class SupportsToplist(Protocol):
    @abstractmethod
    def toplist_list(self) -> List[BriefPlaylistModel]:
        """List all toplist(排行榜)."""

    @abstractmethod
    def toplist_get(self, toplist_id) -> PlaylistModel:
        """Get toplist details.

        For some providers, the toplist model(schema) may be different from the
        PlaylistModel. They should think about a way to solve this. For example,
        turn the identifier into `toplist_{id}` and do some hack in playlist_get API.
        """


#
# Protocols for recommendation.
#
@runtime_checkable
class SupportsRecListDailySongs(Protocol):
    @abstractmethod
    def rec_list_daily_songs(self) -> List[SongModel]:
        pass


@runtime_checkable
class SupportsRecACollectionOfSongs(Protocol):
    @abstractmethod
    def rec_a_collection_of_songs(self) -> Collection:
        """
        For example, providers may provider a list of songs,
        and the title looks like “大家都在听” / “红心歌曲”.

        For different user, this API may return different result.
        """


@runtime_checkable
class SupportsRecACollectionOfVideos(Protocol):
    @abstractmethod
    def rec_a_collection_of_videos(self) -> Collection:
        """
        For example, providers may recommend a list of videos.
        For different user, this API may return different result.
        This API MAY return different result at different time.
        """


@runtime_checkable
class SupportsRecListDailyPlaylists(Protocol):
    @abstractmethod
    def rec_list_daily_playlists(self) -> List[PlaylistModel]:
        pass


@runtime_checkable
class SupportsRecListDailyAlbums(Protocol):
    @abstractmethod
    def rec_list_daily_albums(self) -> List[AlbumModel]:
        pass



from typing import TYPE_CHECKING

from PyQt5.QtCore import QRect
from PyQt5.QtWidgets import QMenu, QAction
from PyQt5.QtGui import QPainter, QIcon, QPalette, QContextMenuEvent

from feeluown.library import UserModel, SupportsCurrentUser, Provider, \
    SupportsCurrentUserChanged
from feeluown.library import reverse
from feeluown.utils.aio import run_afn, run_fn
from feeluown.gui.provider_ui import UISupportsLoginOrGoHome, ProviderUiItem, \
    UISupportsLoginEvent, UISupportsContextMenuAddItems
from feeluown.gui.widgets import SelfPaintAbstractIconTextButton
from feeluown.gui.drawers import SizedPixmapDrawer, AvatarIconDrawer
from feeluown.gui.helpers import painter_save

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class Avatar(SelfPaintAbstractIconTextButton):
    """
    When no provider is selected, click this button will popup a menu,
    and let user select a provider. When a provider is selected, click this
    button will trigger `UISupportsLoginOrGoHome`. If the provider has
    a current user, this tries to show the user avatar.
    """

    def __init__(self, app: 'GuiApp', *args, **kwargs):
        super().__init__('未登录', *args, **kwargs)

        self._app = app
        self._logging_state = {}
        self._avatar_drawer = None
        # In order to make the avatar/icon align to the left edge,
        # translate the painter to -self._padding and set different padding
        # for avatar-icon and avatar-image.
        self._avatar_padding = self._padding // 2
        # Leave 1px to draw line itself.
        # Theoretically, the line itself costs 1.5px and only 0.75px is needed.
        self._translate_x = 1 - self._padding
        self._avatar_translate_x = -self._avatar_padding
        self._icon_drawer = AvatarIconDrawer(self.height(), self._padding)
        self.clicked.connect(self.on_clicked)
        self.setToolTip('点击切换平台')

        self._app.library.provider_added.connect(self.on_provider_added)

    def on_clicked(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        if pvd_ui is not None:
            if isinstance(pvd_ui, UISupportsLoginOrGoHome):
                pvd_ui.login_or_go_home()
            return

        item = self._app.current_pvd_ui_mgr.get_item()
        if item is not None:
            item.clicked.emit()
        else:
            pos = self.cursor().pos()
            e = QContextMenuEvent(QContextMenuEvent.Mouse, pos, pos)
            self.contextMenuEvent(e)

    def on_provider_added(self, provider: Provider):
        if isinstance(provider, SupportsCurrentUserChanged):
            provider.current_user_changed.connect(
                self.create_provider_current_user_changed_cb(provider), weak=False)

    def create_provider_current_user_changed_cb(self, provider: Provider):
        def cb(user: UserModel):
            if user is not None:
                self._logging_state[provider.identifier] = user.name
            else:
                self._logging_state.pop(provider.identifier, None)
            if not self._app.current_pvd_ui_mgr.get():
                if self._logging_state:
                    self._text = '部分已登录'
                else:
                    self._text = '未登录'
                self.setToolTip(self._text)  # refresh tooltip
        return cb

    def contextMenuEvent(self, e) -> None:
        # pylint: disable=unnecessary-direct-lambda-call
        menu = QMenu()

        # Let provider UI add custom actions if supported
        current_pvd_ui = self._app.current_pvd_ui_mgr.get()
        if (
            current_pvd_ui is not None and
            isinstance(current_pvd_ui, UISupportsContextMenuAddItems)
        ):
            menu.addSection(current_pvd_ui.provider.name)
            current_pvd_ui.context_menu_add_items(menu)

        # Create a submenu for "切换账号"
        menu.addSection('切换账号')
        switch_account_menu = menu

        for item in self._app.pvd_uimgr.list_items():
            action = QAction(QIcon(item.colorful_svg or 'icons:feeluown.png'),
                             item.text,
                             parent=switch_account_menu)
            action.setToolTip(item.desc)
            action.triggered.connect(
                (lambda item: lambda _: self.on_provider_selected(item))(item))
            switch_account_menu.addAction(action)

        for pvd_ui in self._app.pvd_ui_mgr.list_all():
            action = QAction(QIcon(pvd_ui.get_colorful_svg() or 'icons:feeluown.png'),
                             pvd_ui.provider.meta.name,
                             parent=switch_account_menu)
            action.triggered.connect(
                (lambda pvd_ui: lambda _: self.on_pvd_ui_selected(pvd_ui))(pvd_ui))
            switch_account_menu.addAction(action)

        # Show menu
        menu.exec_(e.globalPos())

    def on_provider_ui_login_event(self, provider_ui, event):
        current_pvd_ui = self._app.current_pvd_ui_mgr.get()
        if current_pvd_ui == provider_ui and event == 2:
            return
        if event in (1, 2):
            run_afn(self.show_pvd_ui_current_user)
            run_afn(
                self._app.ui.sidebar.show_provider_current_user_playlists,
                provider_ui.provider
            )

    def on_pvd_ui_selected(self, pvd_ui):
        if isinstance(pvd_ui, UISupportsLoginEvent):
            pvd_ui.login_event.connect(self.on_provider_ui_login_event)
        if isinstance(pvd_ui, UISupportsLoginOrGoHome):
            pvd_ui.login_or_go_home()
        # Set current provider ui at the very last.
        # Must not set it before handling login event.
        self._app.current_pvd_ui_mgr.set(pvd_ui)

    def on_provider_selected(self, provider: ProviderUiItem):
        self._app.current_pvd_ui_mgr.set_item(provider)
        self.setToolTip(provider.text + '\n\n' + provider.desc)
        provider.clicked.emit()
        run_afn(self.show_provider_current_user)

    async def show_pvd_ui_current_user(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        name = pvd_ui.provider.meta.identifier
        user = await self._show_provider_current_user(name)
        if user is not None:
            self.setToolTip(f'{user.name} ({pvd_ui.provider.meta.name})')

    async def show_provider_current_user(self):
        item = self._app.current_pvd_ui_mgr.get_item()
        user = await self._show_provider_current_user(item.name)
        if user is not None:
            self.setToolTip(f'{user.name} ({item.text})')

    async def _show_provider_current_user(self, source):
        self.setToolTip('')
        self._avatar_drawer = None
        provider = self._app.library.get(source)
        assert provider is not None
        user = None
        if isinstance(provider, SupportsCurrentUser):
            user = await run_fn(provider.get_current_user_or_none)

        if user is None:
            self._text = '未登录'
            return None
        if isinstance(user, UserModel) and user.avatar_url:
            self._text = user.name
            img_data = await run_afn(self._app.img_mgr.get, user.avatar_url,
                                     reverse(user))
            if img_data:
                p = self._avatar_padding
                w = self.height() - 2 * p
                rect = QRect(p, p, w, w)
                self._avatar_drawer = SizedPixmapDrawer.from_img_data(
                    img_data, rect, radius=0.5)
        return user

    def paint_border_bg_when_hover(self, *_, **__):
        pass

    def setToolTip(self, text):
        notes = f"后台已登录：{','.join(self._logging_state.keys()) or '无'}"
        super().setToolTip(text + '\n\n' + notes)

    def draw_text(self, painter):
        with painter_save(painter):
            if not self._avatar_drawer:
                painter.translate(self._translate_x, 0)
            super().draw_text(painter)

    def draw_icon(self, painter: QPainter):
        if self._avatar_drawer:
            with painter_save(painter):
                painter.translate(self._avatar_translate_x, 0)
                self._avatar_drawer.draw(painter)
        else:
            with painter_save(painter):
                painter.translate(self._translate_x, 0)
                # If a provider is selected, draw a highlight circle.
                if self._app.current_pvd_ui_mgr.get_either() is not None:
                    self._icon_drawer.fg_color = self.palette().color(QPalette.Highlight)
                self._icon_drawer.draw(painter)


if __name__ == '__main__':
    from unittest.mock import MagicMock
    from feeluown.gui.debug import simple_layout, mock_app

    length = 400

    with simple_layout() as layout, mock_app() as mockapp:
        mockapp.pvd_uimgr.list_items = MagicMock(return_value=[
            ProviderUiItem(
                '',
                'Hello World',
                '',
                'Hello World',
            ),
            ProviderUiItem(
                '',
                'Hello PyQt5',
                '',
                'Hello PyQt5',
            )
        ])
        layout.addWidget(Avatar(mockapp, height=length))



from typing import TYPE_CHECKING

from PyQt5.QtCore import QTimer, QRect, Qt
from PyQt5.QtGui import QPainter, QPalette, QColor
from PyQt5.QtWidgets import QLabel, QSizePolicy, QMenu, QVBoxLayout, QWidget

from feeluown.player import PlaylistPlayModelStage
from feeluown.library import fmt_artists_names
from feeluown.gui.components import SongMenuInitializer
from feeluown.gui.helpers import elided_text

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class LineSongLabel(QLabel):
    """Show song info in one line (with limited width)."""

    default_text = '...'

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(text=self.default_text, parent=parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self._app = app

        # TODO: we can create a label class that roll the text when
        # the text is longer than the label width
        self._timer = QTimer()
        self._txt = self._raw_text = self.default_text
        self._text_rect = self.fontMetrics().boundingRect(self._raw_text)
        # text's position, keep changing to make text roll
        self._pos = 0
        self._timer.timeout.connect(self.change_text_position)

        self._app.player.metadata_changed.connect(
            self.on_metadata_changed, aioqueue=True
        )
        self._app.playlist.play_model_stage_changed.connect(
            self.on_play_model_stage_changed, aioqueue=True
        )

    def on_metadata_changed(self, metadata):
        if not metadata:
            self.setText('...')
            return

        # Set main text.
        text = metadata.get('title', '')
        if text:
            artists = metadata.get('artists', [])
            if artists:
                # FIXME: use _get_artists_name
                text += f" • {','.join(artists)}"
        self.setText(text)

    def on_play_model_stage_changed(self, stage):
        if stage == PlaylistPlayModelStage.prepare_media:
            self.setText('正在获取歌曲播放链接...')
        elif stage == PlaylistPlayModelStage.find_standby_by_mv:
            self.setText('正在获取音乐的视频播放链接...')
        elif stage == PlaylistPlayModelStage.find_standby:
            self.setText('尝试寻找备用播放链接...')
        elif stage == PlaylistPlayModelStage.prepare_metadata:
            self.setText('尝试获取完整的歌曲元信息...')
        elif stage == PlaylistPlayModelStage.load_media:
            self.setText('正在加载歌曲资源...')

    def change_text_position(self):
        if not self.parent().isVisible():  # type: ignore
            self._timer.stop()
            self._pos = 0
            return
        if self._text_rect.width() + self._pos > 0:
            # control the speed of rolling
            self._pos -= 5
        else:
            self._pos = self.width()
        self.update()

    def setText(self, text):
        self._txt = self._raw_text = text
        self._text_rect = self.fontMetrics().boundingRect(self._raw_text)
        self._pos = 0
        self.update()

    def enterEvent(self, event):
        # we do not compare text_rect with self_rect here because of
        # https://github.com/feeluown/FeelUOwn/pull/425#discussion_r536817226
        # TODO: find out why
        if self._txt != self._raw_text:
            # decrease to make rolling more fluent
            self._timer.start(150)

    def leaveEvent(self, event):
        self._timer.stop()
        self._pos = 0
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setFont(self.font())
        painter.setPen(self.palette().color(QPalette.Text))

        if self._timer.isActive():
            self._txt = self._raw_text
        else:
            self._txt = self.fontMetrics().elidedText(
                self._raw_text, Qt.ElideRight, self.width()
            )

        painter.drawText(
            QRect(self._pos, 0,
                  self.width() - self._pos, self.height()),
            Qt.AlignLeft | Qt.AlignVCenter, self._txt
        )  # type: ignore[call-overload]

    def contextMenuEvent(self, e):
        song = self._app.playlist.current_song
        if song is None:
            return

        menu = QMenu()
        SongMenuInitializer(self._app, song).apply(menu)
        menu.exec(e.globalPos())


class TwoLineSongLabel(QWidget):
    default_text = '...'

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)
        self._app = app

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self._title_label = QLabel()
        self._subtitle_label = QLabel()

        palette = self._subtitle_label.palette()
        palette.setColor(QPalette.Text, QColor('grey'))
        palette.setColor(QPalette.Foreground, QColor('Grey'))
        self._subtitle_label.setPalette(palette)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(5)
        self._layout.addWidget(self._title_label)
        self._layout.addWidget(self._subtitle_label)
        self._app.player.metadata_changed.connect(
            self.on_metadata_changed, aioqueue=True
        )

        font = self.font()
        font.setPixelSize(25)
        font.setBold(True)
        self._title_label.setFont(font)
        font.setPixelSize(20)
        font.setBold(False)
        self._subtitle_label.setFont(font)

    def on_metadata_changed(self, metadata):
        if not metadata:
            self._title_label.setText('...')
            return

        # Set main text.
        title = metadata.get('title', '')
        if title:
            artists = metadata.get('artists', [])
            if artists:
                # FIXME: use _get_artists_name
                subtitle = fmt_artists_names(artists)
                self._subtitle_label.setText(
                    elided_text(
                        subtitle, self._subtitle_label.width(),
                        self._subtitle_label.font()
                    )
                )
        self._title_label.setText(
            elided_text(title, self._title_label.width(), self._title_label.font())
        )



import re

from .cmd import Cmd
from .excs import HandlerException
from .base import AbstractHandler


class SubHandler(AbstractHandler):
    cmds = ('sub', )

    def handle(self, cmd: Cmd):
        return self.handle_sub(*cmd.args)

    def handle_sub(self, *topics):
        pubsub_gateway = self._app.pubsub_gateway

        matched_topics = []
        for topic in topics:
            # Keep backward compatibility.
            if topic.startswith('topic.'):
                topic = topic[6:]

            p = re.compile(rf'{topic}')
            for topic_name in pubsub_gateway.topics:
                m = p.match(topic_name)
                if m is not None and m.group(0) == topic_name:
                    matched_topics.append(topic_name)

            if matched_topics:
                for each in matched_topics:
                    pubsub_gateway.link(each, self.session)
            else:
                raise HandlerException(f"{topic}: topic not found")



import logging
from typing import Callable, Optional, List, Tuple, Dict

from feeluown.utils.dispatch import Signal


logger = logging.getLogger(__name__)


class SignalConnector:
    def __init__(self, symbol: str):
        self._signal: Optional[Signal] = None
        self.symbol = symbol

        self._slot_list: List[Tuple[Callable, Dict]] = []
        self._slot_symbol_list: List[Tuple[str, Dict]] = []

    def bind_signal(self, signal: Signal):
        self._signal = signal

    def connect(self):
        """Connect all slots.
        """
        if self._signal is None:
            raise RuntimeError("no signal is bound")

        # Connect slot which are not symbol.
        # These slots are connected directly.
        for slot, kwargs in self._slot_list:
            self._signal.connect(slot, **kwargs)
        self._slot_list.clear()

        # Connect slots which are symbol currently.
        self._signal.connect(self.slot_symbols_delegate, weak=False)

    def connect_slot(self, slot: Callable, **kwargs):
        if self._signal is not None:
            # If signal is already bound, the slot is connected immediately.
            self._signal.connect(slot, **kwargs)
        else:
            self._slot_list.append((slot, kwargs))

    def connect_slot_symbol(self, slot_symbol: str, **kwargs):
        if (slot_symbol, kwargs) not in self._slot_symbol_list:
            self._slot_symbol_list.append((slot_symbol, kwargs))

    def disconnect_slot_symbol(self, slot_symbol):
        for i, (symbol, _) in enumerate(self._slot_symbol_list):
            if symbol == slot_symbol:
                self._slot_symbol_list.pop(i)

    def disconnect_slot(self, slot: Callable):
        if self._signal is not None:
            self._signal.disconnect(slot)
        else:
            for i, (s, _) in enumerate(self._slot_list):
                if s == slot:
                    self._slot_list.pop(i)
                    break

    def slot_symbols_delegate(self, *args):
        """
        A delegate invoke the slots for the signal.

        Signal.emit => self.slot_symbols_delegate => slots
        """
        for slot_symbol, kwargs in self._slot_symbol_list:
            func = fuoexec_F(slot_symbol)
            # FIXME: Duplicate code. The logic has been implemented in Signal.emit.
            if kwargs.get('aioqueue'):
                if Signal.has_aio_support:
                    Signal.aioqueue.sync_q.put_nowait((func, args))  # type: ignore
                else:
                    logger.warning(
                        'No aio support is available, a slot is ignored.')
            else:
                try:
                    func(*args)
                except:  # noqa, pylint: disable=bare-except
                    logger.exception('error during calling slot:%s')


class SignalManager:

    def __init__(self):
        self.initialized = False
        self._app = None

        self.signal_connectors: List[SignalConnector] = []

    def initialize(self, app):
        """
        Find each signal by signal_symbol and connect slots for them.
        """
        if self.initialized:
            raise RuntimeError('signals slots manager already initialized')

        self._app = app
        for connector in self.signal_connectors:
            self._init_sc(connector)
        self.initialized = True

    def add(self, signal_symbol: str, slot: Callable, use_symbol: bool, **kwargs):
        """Add one slot for the signal.

        :param slot: The function or it's symbol.
        """
        sc = self._get_or_create_sc(signal_symbol)
        if use_symbol is True:
            sc.connect_slot_symbol(fuoexec_S(slot), **kwargs)
        else:
            sc.connect_slot(slot, **kwargs)

    def remove(self, signal_symbol: str, slot: Callable, use_symbol: bool):
        """Remove one slot for signal.

        If slot is not connected, this does nothing.
        """
        signal_connector = self._get_or_create_sc(signal_symbol)
        if use_symbol is True:
            signal_connector.disconnect_slot_symbol(fuoexec_S(slot))
        else:
            signal_connector.disconnect_slot(slot)

    def _get_or_create_sc(self, signal_symbol) -> SignalConnector:
        """Get or create signal connector."""
        for sc in self.signal_connectors:
            if sc.symbol == signal_symbol:
                signal_connector = sc
                break
        else:
            signal_connector = SignalConnector(signal_symbol)
            if self.initialized:
                self._init_sc(signal_connector)
            self.signal_connectors.append(signal_connector)
        return signal_connector

    def _init_sc(self, sc):
        # pylint: disable=eval-used
        signal = eval(sc.symbol, {'app': self._app})
        sc.bind_signal(signal)
        sc.connect()


signal_mgr: SignalManager = SignalManager()


# pylint: disable=wrong-import-position, cyclic-import
from .fuoexec import fuoexec_S, fuoexec_F  # noqa



import asyncio
import socket

from openai import AsyncOpenAI

from feeluown.utils.aio import run_afn


async def a_handle_stream(stream):
    rsock, wsock = socket.socketpair()
    rr, rw = await asyncio.open_connection(sock=rsock)
    _, ww = await asyncio.open_connection(sock=wsock)

    async def write_task():
        async for chunk in stream:
            # When stream_options={'include_usage': True},
            # the last chunk.choices is empty.
            if chunk.choices:
                content = chunk.choices[0].delta.content or ''
                ww.write(content.encode('utf-8'))

        ww.write_eof()
        await ww.drain()
        ww.close()
        await ww.wait_closed()
        try:
            return chunk  # chunk may be undefined
        except NameError:
            return

    task = run_afn(write_task)
    return rr, rw, task


class AI:
    def __init__(self, base_url, api_key, model):
        self.base_url = base_url
        self.api_key = api_key
        self.model = model

    def get_async_client(self):
        return AsyncOpenAI(
            base_url=self.base_url,
            api_key=self.api_key,
        )



from typing import TYPE_CHECKING

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout
from feeluown.library.provider_protocol import SupportsToplist

from feeluown.utils.reader import create_reader
from feeluown.utils.aio import run_fn
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.img_card_list import (
    PlaylistCardListView, PlaylistCardListModel, PlaylistFilterProxyModel,
    PlaylistCardListDelegate,
)

from feeluown.library import (
    SupportsRecListDailyPlaylists, SupportsRecListDailySongs,
    SupportsCurrentUserDislikeSongsReader,
)

from feeluown.gui.widgets import CalendarButton, RankButton, EmojiButton
from feeluown.gui.helpers import fetch_cover_wrapper


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


async def render(req, **kwargs):
    app: 'GuiApp' = req.ctx['app']

    view = View(app)
    app.ui.right_panel.set_body(view)
    await view.render()


class View(QWidget):
    def __init__(self, app: 'GuiApp'):
        super().__init__(parent=None)
        self._app = app

        self.header_title = LargeHeader()
        self.header_playlist_list = MidHeader()
        self.playlist_list_view = PlaylistCardListView(fixed_row_count=1)
        self.playlist_list_view.setItemDelegate(
            PlaylistCardListDelegate(self.playlist_list_view,
                                     card_min_width=100,))
        self.daily_songs_btn = CalendarButton('每日推荐', parent=self)
        self.rank_btn = RankButton(parent=self)
        # FIXME: design a new button for dislike
        self.dislike_btn = EmojiButton('🚫', '音乐黑名单', parent=self)
        self.daily_songs_btn.setMinimumWidth(150)
        self.rank_btn.setMinimumWidth(150)
        self.dislike_btn.setMinimumWidth(150)

        self.header_title.setText('发现音乐')
        self.header_playlist_list.setText('个性化推荐')

        self._layout = QVBoxLayout(self)
        self._setup_ui()

        self.playlist_list_view.show_playlist_needed.connect(
            lambda model: self._app.browser.goto(model=model))
        self.daily_songs_btn.clicked.connect(
            lambda: self._app.browser.goto(page='/rec/daily_songs'))
        self.rank_btn.clicked.connect(
            lambda: self._app.browser.goto(page='/toplist'))
        self.dislike_btn.clicked.connect(
            lambda: self._app.browser.goto(page='/my_dislike'))

    def _setup_ui(self):
        self._h_layout = QHBoxLayout()
        self._h_layout.addWidget(self.daily_songs_btn)
        self._h_layout.addSpacing(10)
        self._h_layout.addWidget(self.rank_btn)
        self._h_layout.addSpacing(10)
        self._h_layout.addWidget(self.dislike_btn)
        self._h_layout.addStretch(0)

        self._layout.setContentsMargins(20, 10, 20, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.header_title)
        self._layout.addSpacing(10)
        self._layout.addLayout(self._h_layout)
        self._layout.addSpacing(30)
        self._layout.addWidget(self.header_playlist_list)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.playlist_list_view)
        self._layout.addStretch(0)

    async def render(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        if pvd_ui is None:
            self._app.show_msg('wtf!')
            return

        provider = pvd_ui.provider
        if isinstance(provider, SupportsRecListDailyPlaylists):
            playlists = await run_fn(provider.rec_list_daily_playlists)
            model = PlaylistCardListModel(
                create_reader(playlists),
                fetch_cover_wrapper(self._app),
                {p.identifier: p.name for p in self._app.library.list()})
            filter_model = PlaylistFilterProxyModel()
            filter_model.setSourceModel(model)
            self.playlist_list_view.setModel(filter_model)

        self.daily_songs_btn.setEnabled(isinstance(provider, SupportsRecListDailySongs))
        self.rank_btn.setEnabled(isinstance(provider, SupportsToplist))
        self.dislike_btn.setEnabled(
            isinstance(provider, SupportsCurrentUserDislikeSongsReader))



import os
import sys

from PyQt5.QtCore import Qt, QDir
from PyQt5.QtGui import QIcon, QPixmap, QGuiApplication
from PyQt5.QtWidgets import QApplication, QWidget

from feeluown.gui.browser import Browser
from feeluown.gui.hotkey import HotkeyManager
from feeluown.gui.image import ImgManager
from feeluown.gui.theme import ThemeManager
from feeluown.gui.tips import TipsManager
from feeluown.gui.watch import WatchManager
from feeluown.gui.ui import Ui
from feeluown.gui.tray import Tray
from feeluown.gui.provider_ui import ProviderUiManager, CurrentProviderUiManager
from feeluown.gui.uimodels.playlist import PlaylistUiManager
from feeluown.gui.uimodels.my_music import MyMusicUiManager

from .app import App


class GuiApp(App, QWidget):

    def __init__(self, *args, **kwargs):
        config = args[1]
        pkg_root_dir = os.path.join(os.path.dirname(__file__), '..')
        icons_dir = os.path.join(pkg_root_dir, 'gui/assets/icons')
        QDir.addSearchPath('icons', icons_dir)
        QGuiApplication.setWindowIcon(QIcon(QPixmap('icons:feeluown.png')))
        # Set desktopFileName so that the window icon is properly shown under wayland.
        # I don't know if this setting brings other benefits or not.
        # https://github.com/pyfa-org/Pyfa/issues/1607#issuecomment-392099878
        QApplication.setDesktopFileName('FeelUOwn')
        QApplication.instance().setQuitOnLastWindowClosed(not config.ENABLE_TRAY)
        QApplication.instance().setApplicationName('FeelUOwn')

        # Note that QApplication.setFont only works for those widgets that created
        # after `QApplication.exec` (tested on macOS).
        if sys.platform == 'win32':
            font = QApplication.font()
            # By default, it uses SimSun(宋体) on windows, which is a little ugly.
            # "Segoe UI Symbol" is used to render charactor symbols.
            # "Microsoft Yahei" is used to render chinese (and english).
            # Choose a default sans-serif font when the first two fonts do not work,
            font.setFamilies(['Segoe UI Symbol', 'Microsoft YaHei', 'sans-serif'])

            # When a HiDPI screen is used, users need to set both font DPI and
            # screen scale factor to make it working properly when pointSize is used.
            # It's hard for most users to set them right.
            # When using pixelSize, users only need to set screen scale factor.
            # In other words, only QT_AUTO_SCREEN_SCALE_FACTOR=1 is needed to set
            # and feeluown can works properly in HiDPI environment.
            #
            # Based on past experience, 13px is the default font size for all platform,
            # including windows, linux and macOS.
            font.setPixelSize(13)
            QApplication.setFont(font)
        elif sys.platform == 'darwin':
            font = QApplication.font()
            # The default font can not show chinese in bold.
            font.setFamilies(['arial', 'Hiragino Sans GB', 'sans-serif'])
            QApplication.setFont(font)

        QWidget.__init__(self)
        App.__init__(self, *args, **kwargs)

        GuiApp.__q_app = QApplication.instance()

        self.setObjectName('app')

        # GUI 的一些辅助管理模块
        self.theme_mgr = ThemeManager(self, parent=self)
        self.tips_mgr = TipsManager(self)
        self.hotkey_mgr = HotkeyManager(self)
        self.img_mgr = ImgManager(self)
        self.watch_mgr = WatchManager(self)

        # GUI 组件的数据管理模块
        self.pvd_ui_mgr = self.pvd_uimgr = ProviderUiManager(self)
        self.current_pvd_ui_mgr = CurrentProviderUiManager(self)
        self.pl_uimgr = PlaylistUiManager(self)
        self.mymusic_uimgr = MyMusicUiManager(self)

        self.browser = Browser(self)
        self.ui = Ui(self)
        if self.config.ENABLE_TRAY:
            self.tray = Tray(self)
        self.show_msg = self.ui._message_line.show_msg

    def initialize(self):
        super().initialize()

        self.hotkey_mgr.initialize()
        self.theme_mgr.initialize()
        if self.config.ENABLE_TRAY:
            self.tray.initialize()
            self.tray.show()
        self.watch_mgr.initialize()
        self.browser.initialize()
        QApplication.instance().aboutToQuit.connect(self.about_to_exit)

    def run(self):
        self.show()
        super().run()

    def apply_state(self, state):
        super().apply_state(state)
        coll_library = self.coll_mgr.get_coll_library()
        self.browser.goto(page=f'/colls/{coll_library.identifier}')

        gui = state.get('gui', {})
        lyric = gui.get('lyric', {})
        local_storage = gui.get('browser', {}).get('local_storage', {})
        self.browser.local_storage = local_storage
        self.ui.lyric_window.apply_state(lyric)

    def dump_state(self):
        state = super().dump_state()
        state['gui'] = {
            'lyric': self.ui.lyric_window.dump_state(),
            'browser': {
                'local_storage': self.browser.local_storage
            }
        }
        return state

    def closeEvent(self, _):
        if not self.config.ENABLE_TRAY:
            self.exit()

    def mouseReleaseEvent(self, e):
        if not self.rect().contains(e.pos()):
            return
        if e.button() == Qt.BackButton:
            self.browser.back()
        elif e.button() == Qt.ForwardButton:
            self.browser.forward()

    def exit_player(self):
        # Destroy GL context or mpv renderer
        self.ui.mpv_widget.shutdown()
        super().exit_player()

    def about_to_exit(self):
        super().about_to_exit()
        QApplication.instance().aboutToQuit.disconnect(self.about_to_exit)

    def exit(self):
        QApplication.exit()



import logging
from typing import List, Union
from urllib.parse import urlparse

from PyQt5.QtCore import pyqtSignal, QUrl, QRect
from PyQt5.QtNetwork import QNetworkCookie
from PyQt5.QtWebEngineCore import QWebEngineCookieStore
from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEngineProfile, \
    QWebEnginePage
from PyQt5.QtWidgets import QApplication, QDesktopWidget


logger = logging.getLogger(__name__)


class NoOutputWebPage(QWebEnginePage):
    def javaScriptConsoleMessage(self, level, msg, line, sourceID):
        pass


class WebLoginView(QWebEngineView):
    succeed = pyqtSignal(dict)

    def __init__(self, uri: str, required_cookies: List[Union[str, List[str]]],
                 parent=None):
        """
        This utility class is used for automating the cookie login process.
        Example: QtWebEngine('https://y.xx.com', ['xxmusic_key'])

        :param uri: The initial login URL.
        :param required_cookies: List of required Cookie keys. When all are obtained,
            the succeed signal will be emitted and the window will close automatically.
            You can also use a list of lists, for example:
               [['xx_key', 'xx_wxuin'], ['xx_key', 'xx_uin]]
        :param parent:
        """
        super().__init__(parent)
        self.set_pos()
        self.init_uri = uri
        profile = QWebEngineProfile.defaultProfile()
        cookie_store: QWebEngineCookieStore = profile.cookieStore()
        cookie_store.deleteAllCookies()
        cookie_store.cookieAdded.connect(self.cookie_added)
        cookie_store.cookieRemoved.connect(self.cookie_removed)
        self.saved_cookies = dict()  # type: ignore
        self.required_cookies_options = []
        if required_cookies:
            if isinstance(required_cookies[0], list):
                for option in required_cookies:
                    self.required_cookies_options.append(option)
            else:
                self.required_cookies_options.append(required_cookies)
        self.setPage(NoOutputWebPage(self))
        self.load(QUrl(uri))

    def set_pos(self):
        desktop: QDesktopWidget = QApplication.desktop()
        screen = desktop.screenNumber(QApplication.desktop().cursor().pos())
        geo: QRect = desktop.availableGeometry(screen)
        self.setFixedWidth(int(geo.width() / 1.5))
        self.setFixedHeight(int(geo.height() / 1.5))
        frame: QRect = self.frameGeometry()
        frame.moveCenter(geo.center())
        self.move(frame.topLeft())

    def check_cookie_domain(self, cookie: QNetworkCookie):
        """
        检查 Cookie 的域名
        :param cookie: Cookie
        :type cookie: QNetworkCookie
        :return: 是否匹配域名
        :rtype: bool
        """
        cookie_domain = cookie.domain().lstrip('.')
        urisegs = urlparse(self.init_uri)
        if urisegs.hostname is not None:
            return urisegs.hostname.endswith(cookie_domain)
        return False

    def cookie_added(self, cookie: QNetworkCookie):
        if self.check_cookie_domain(cookie):
            name = cookie.name().data().decode()
            value = cookie.value().data().decode()
            self.saved_cookies[name] = value
            for required_cookies in self.required_cookies_options:
                ok = True
                one_ok = False
                for _name in required_cookies:
                    if _name not in self.saved_cookies:
                        ok = False
                        if one_ok is True:  # log for debugging
                            logger.debug(
                                f"not enough cookies for login, "
                                f"required: {required_cookies}, "
                                f"not satisfied: {_name}"
                            )
                        break
                    one_ok = True

                if ok is True:
                    logger.debug(
                        f"enough cookies for login, "
                        f"required: {self.required_cookies_options}, "
                        f"acture: {self.saved_cookies}"
                    )
                    self.succeed.emit(self.saved_cookies)
                    break

    def cookie_removed(self, cookie: QNetworkCookie):
        if cookie.name().data().decode() in self.saved_cookies.keys():
            self.saved_cookies.pop(cookie.name().data().decode())


if __name__ == '__main__':
    app = QApplication([])
    widget = WebLoginView("xxx", ['xxx_key'])
    widget.show()
    app.exec()



import asyncio
import sys

from qasync import QEventLoop, QThreadExecutor
from PyQt5.QtWidgets import QApplication

try:
    from qasync import DefaultQEventLoopPolicy
except ImportError:
    # From qasync>=0.28, the class DefaultQEventLoopPolicy is removed
    # when python version >= 3.12
    class DefaultQEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
        def new_event_loop(self):
            return QEventLoop(QApplication.instance() or QApplication(sys.argv))


__all__ = (
    'QEventLoop',
    'QThreadExecutor',
    'DefaultQEventLoopPolicy'
)



# -*- coding: utf-8 -*-
import os
import logging

from feeluown.utils import aio  # noqa
from .provider import provider  # noqa
from .provider_ui import LocalProviderUi

DEFAULT_MUSIC_FOLDER = os.path.expanduser('~') + '/Music'
DEFAULT_MUSIC_EXTS = ['mp3', 'ogg', 'wma', 'm4a', 'm4v', 'mp4', 'flac', 'ape', 'wav']

__alias__ = '本地音乐'
__feeluown_version__ = '1.1.0'
__version__ = '0.1a0'
__desc__ = '本地音乐'

logger = logging.getLogger(__name__)


def init_config(config):
    config.deffield('MUSIC_FOLDERS',
                    type_=list,
                    default=[DEFAULT_MUSIC_FOLDER],
                    desc='支持的音乐文件夹列表')
    config.deffield('MUSIC_FORMATS',
                    type_=list,
                    default=DEFAULT_MUSIC_EXTS,
                    desc='支持的音乐格式列表')
    config.deffield('CORE_LANGUAGE', type_=str, default='auto', desc='默认显示的语言')
    config.deffield('IDENTIFIER_DELIMITER',
                    type_=str,
                    default='',
                    desc='生成identifier时的连接符')
    config.deffield('EXPAND_ARTIST_SONGS',
                    type_=bool,
                    default=False,
                    desc='将专辑艺术家的专辑中歌曲加入到该艺术家的歌曲中')
    config.deffield('ARTIST_SPLITTER', type_=list, default=[',', '&'], desc='歌曲艺术家的分隔符')
    config.deffield('ARTIST_SPLITTER_IGNORANCE',
                    type_=list,
                    default=None,
                    desc='对艺术家信息使用分隔符时需要进行保护的字符串')
    config.deffield('SPLIT_ALBUM_ARTIST_NAME',
                    type_=bool,
                    default=False,
                    desc='支持使用分隔符分隔专辑艺术家')


async def autoload(app):
    await aio.run_fn(provider.scan, app.config.local, app.config.local.MUSIC_FOLDERS)

    app.show_msg('本地音乐扫描完毕')


def enable(app):
    logger.info('Register provider: %s', provider)
    app.library.register(provider)
    provider.initialize(app)

    app.started.connect(lambda *args: aio.create_task(autoload(*args)),
                        weak=False,
                        aioqueue=False)
    if app.mode & app.GuiMode:
        provider_ui = LocalProviderUi(app)
        app.pvd_ui_mgr.register(provider_ui)


def disable(app):
    logger.info('唔，不要禁用我')



from bisect import bisect
from string import Formatter


widths = [
    (0, 1),
    (126, 1), (159, 0), (687, 1), (710, 0), (711, 1),
    (727, 0), (733, 1), (879, 0), (1154, 1), (1161, 0),
    (4347, 1), (4447, 2), (7467, 1), (7521, 0), (8369, 1),
    (8426, 0), (9000, 1), (9002, 2), (11021, 1), (12350, 2),
    (12351, 1), (12438, 2), (12442, 0), (19893, 2), (19967, 1),
    (55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
    (65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2),
    (120831, 1), (262141, 2), (1114109, 1),
]
points = [w[0] for w in widths]


def char_len(c):
    ord_code = ord(c)
    if ord_code == 0xe or ord_code == 0xf:
        return 0
    i = bisect(points, ord_code)
    return widths[i][1]


def _fit_text(text, length, filling=True):
    assert 80 >= length >= 5

    text_len = 0
    len_index_map = {}
    for i, c in enumerate(text):
        text_len += char_len(c)
        len_index_map[text_len] = i

    if text_len <= length:
        if filling:
            return text + (length - text_len) * ' '
        return text

    remain = length - 1
    if remain in len_index_map:
        return text[:(len_index_map[remain] + 1)] + '…'
    else:
        return text[:(len_index_map[remain - 1] + 1)] + ' …'


class WideFormatter(Formatter):
    """
    Custom string formatter that handles new format parameters:
    '_': _fit_text(*, filling=False)
    '+': _fit_text(*, filling=True)
    """

    def format(self, format_string, *args, **kwargs):
        return super().vformat(format_string, args, kwargs)

    def format_field(self, value, format_spec):
        if value is None:
            value = 'null'
        fmt_type = format_spec[0] if format_spec else None
        if fmt_type == "_":
            return _fit_text(value, int(format_spec[1:]), filling=False)
        elif fmt_type == "+":
            return _fit_text(value, int(format_spec[1:]), filling=True)
        return super().format_field(value, format_spec)



from PyQt5.QtCore import QMimeData

from feeluown.library import ModelType


model_mimetype_map = {
    ModelType.dummy.value: 'fuo-model/x-dummy',
    ModelType.song.value: 'fuo-model/x-song',
    ModelType.playlist.value: 'fuo-model/x-playlist',
    ModelType.album.value: 'fuo-model/x-album',
    ModelType.artist.value: 'fuo-model/x-artist',
    ModelType.lyric.value: 'fuo-model/x-lyric',
    ModelType.user.value: 'fuo-model/x-user',
}


def get_model_mimetype(model):
    return model_mimetype_map[ModelType(model.meta.model_type).value]


class ModelMimeData(QMimeData):
    def __init__(self, model):
        super().__init__()

        self.model = model
        self._mimetype = get_model_mimetype(model)

    def setData(self, format, model):
        self._model = model

    def formats(self):
        return [self._mimetype]

    def hasFormat(self, format):
        return format == self._mimetype

    def data(self, format):
        if format == self._mimetype:
            return self.model
        return None



import asyncio
import os
import sys
from contextlib import contextmanager
from socket import socket, AF_INET, SOCK_STREAM

from feeluown.consts import CACHE_DIR
from feeluown.library import fmt_artists_names
from feeluown.server import Request, Response
from feeluown.server.dslv2 import unparse
from feeluown.server.server import handle_request
from feeluown.utils import aio


OUTPUT_CACHE_FILEPATH = os.path.join(CACHE_DIR, 'cli.out')


def print_error(*args, **kwargs):
    print('\033[0;31m', end='')
    print(*args, **kwargs)
    print('\033[0m', end='')


cmd_handler_mapping = {}


class Client:
    def __init__(self, sock):
        self.sock = sock

    async def send(self, req) -> Response:
        rfile = self.sock.makefile('rb')
        rfile.readline()  # welcome message
        self.sock.send(bytes(unparse(req) + '\n', 'utf-8'))
        line = rfile.readline().decode('utf-8').strip()
        _, code, length = line.split(' ')
        buf = bytearray()
        while len(buf) < int(length) + 2:
            buf.extend(rfile.readline())
        text = bytes(buf[:-2]).decode('utf-8')
        return Response(ok=code.lower() == 'ok', text=text)

    def close(self):
        self.sock.close()


@contextmanager
def connect():
    sock = socket(AF_INET, SOCK_STREAM)
    sock.connect(('127.0.0.1', 23333))
    client = Client(sock)
    try:
        yield client
    except RuntimeError as e:
        print(e)
    client.close()


class HandlerMeta(type):
    def __new__(cls, name, bases, attrs):
        # pylint: disable=duplicate-code
        # FIXME: duplicate code.
        klass = type.__new__(cls, name, bases, attrs)
        if 'cmds' in attrs:
            cmds = attrs['cmds']
            if isinstance(cmds, str):
                cmd_handler_mapping[cmds] = klass
            else:
                for cmd in cmds:
                    cmd_handler_mapping[cmd] = klass
        return klass


class BaseHandler(metaclass=HandlerMeta):

    def __init__(self, args):
        self.args = args

        options_list = ["format"]
        args_dict = vars(args)
        req_options = {option: args_dict.get(option) for option in options_list
                       if args_dict.get(option, None)}
        self._req = Request(args.cmd, options=req_options)

    def before_request(self):
        """before request hook"""

    def get_req(self):
        return self._req

    def process_resp(self, resp):
        if resp.code == 'OK':
            if resp.text:
                print(resp.text)
        else:
            print_error(resp.text)


class SimpleHandler(BaseHandler):

    cmds = (
        'pause', 'resume', 'stop',
        'next', 'previous',
        'toggle', 'clear', 'status',
    )


class HandlerWithWriteListCache(BaseHandler):
    cmds = ('list', 'search')

    def before_request(self):
        cmd = self.args.cmd
        # Search
        if cmd == 'search':
            self._req.cmd_args = (self.args.keyword, )
            source = self.args.source
            type_ = self.args.type
            cmd_options = {}
            if source:
                # FIXME: v1 does not accept list.
                cmd_options['source'] = source[0]
            if type_:
                cmd_options['type'] = type_
            self._req.cmd_options = cmd_options

    def process_resp(self, resp):
        if resp.code != 'OK' or not resp.text:
            super().process_resp(resp)
            return
        fmt = self._req.options.get('format', 'plain')
        if fmt != 'plain':
            print(resp.text)
            return
        lines = resp.text.split('\n')
        with open(OUTPUT_CACHE_FILEPATH, 'w', encoding='utf-8') as f:
            padding_width = len(str(len(lines)))
            tpl = '{:%dd} {}' % padding_width  # pylint: disable=consider-using-f-string
            for index, line in enumerate(lines):
                print(tpl.format(index, line))
                f.write(f'{line}\n')


class HandlerWithReadListCache(BaseHandler):
    cmds = ('show', 'play', 'remove')

    def before_request(self):
        uri = self.args.uri
        try:
            lineno = int(uri)
        except ValueError:
            pass
        else:
            with open(OUTPUT_CACHE_FILEPATH, encoding='utf-8') as f:
                i = 0
                for line in f:
                    if i == lineno:
                        uri = line.split('#')[0].strip()
                        break
                    i += 1
        self._req.cmd_args = (uri, )


class AddHandler(BaseHandler):
    cmds = ('add', )

    def before_request(self):
        furi_list = []
        if self.args.uri is None:
            for line in sys.stdin:
                furi_list.append(line.strip())
        else:
            furi_list = [self.args.uri]
        self._req.cmd_args = (' '.join(furi_list), )


class ExecHandler(BaseHandler):
    cmds = ('exec', )

    def before_request(self):
        code = self.args.code
        self._req.cmd_args = (code, )


class JsonRPCHandler(BaseHandler):
    cmds = ('jsonrpc', )

    def before_request(self):
        body = self.args.body
        if body is None:
            body = sys.stdin.read()
        self._req.cmd_args = (body, )


class OnceClient:
    def __init__(self, app):
        self._app = app

    async def send(self, req: Request) -> Response:
        app = self._app
        return await handle_request(req, app)


async def dispatch(args, client):
    if '"' in (getattr(args, 'cli', '') or '') \
       or '"' in (getattr(args, 'code', '') or '') \
       or '"' in (getattr(args, 'keyword', '') or ''):
        print_error("command args must not contain charactor '\"'")
        return

    HandlerCls = cmd_handler_mapping[args.cmd]
    handler = HandlerCls(args)
    resp = handler.before_request()
    if resp is None:
        resp = await client.send(handler.get_req())
    handler.process_resp(resp)


async def climain(args):
    """dispatch request"""

    # FIXME: move this code to somewhere else
    if args.cmd == 'genicon':
        # pylint: disable=import-outside-toplevel
        from .install import generate_icon
        generate_icon()
        return

    with connect() as cli:
        await dispatch(args, cli)


async def oncemain(app):

    client = OnceClient(app)
    await dispatch(app.args, client)

    if app.args.cmd == 'play':

        def on_metadata_changed(metadata):
            if not metadata:
                return
            uri = metadata.get('uri', '')
            text = metadata.get('title', '')
            artists = metadata.get('artists', [])
            if artists:
                text += ' - '
                text += fmt_artists_names(artists)
            if uri:
                if text:
                    print(f'Playing: {uri} # {text}')
                else:
                    print(f'Playing: {uri}')
            else:
                print(f'Playing: {text}')

        app.player.metadata_changed.connect(on_metadata_changed, weak=False)
        # Hack: wait for 1s so that the url is fetched and player has set
        # the proper playback.
        await asyncio.sleep(1)
        # mpv wait_for_playback will wait until one song is finished,
        # if we have multiple song to play, this will not work well.
        # pylint: disable=protected-access
        await aio.run_fn(app.player._mpv.wait_for_playback)
        app.exit()
    else:
        app.exit()



import asyncio
import logging
import json
from typing import List, TYPE_CHECKING

from feeluown.ai import a_handle_stream
from feeluown.utils.aio import run_afn, as_completed

if TYPE_CHECKING:
    from feeluown.ai import AI
    from feeluown.library import BriefSongModel

logger = logging.getLogger(__name__)


class AIStandbyMatcher:
    # 调试的使用，可以考虑把 reason 字段也带上。
    #   {"song_id": "xxx", "score": 100, "reason": ""}
    #   {"song_id": "yyy", "score": 95, "reason": ""}
    #
    #   {"song_id": "xxx", "score": 100}
    #   {"song_id": "yyy", "score": 95}
    STANDBY_MATCH_PROMPT = '''\
你是一个音乐播放器助手。你根据“匹配度”帮助用户对搜索候选项进行排序。\
判断匹配度的时候，简体和繁体、英文和中文这些因素都可以忽略，你可以智能转换，然后看是否匹配。\
你要重点考虑原唱这个因素,类似“重制、翻唱、非原创”的歌曲，排序都应该靠后。\
排序完之后，你需要用类似下面的格式来返回，每行一个 JSON，你返回的内容不能包含其它内容

    {"song_id": "xxx", "score": 100}
    {"song_id": "yyy", "score": 95}
'''

    def __init__(self, ai: 'AI', a_prepare_media, min_score, audio_select_policy):
        self.ai = ai
        self._prepare_media = a_prepare_media

        self.min_score = min_score
        self.audio_select_policy = audio_select_policy

        # during runtime
        self.fetch_media_tasks = {}
        self.song_by_source = {}

    async def match(self, song: 'BriefSongModel', standby_list: List['BriefSongModel']):
        sys_msg = {'role': 'system', 'content': self.STANDBY_MATCH_PROMPT}
        user_msg = {
            'role': 'user',
            'content': (f'我搜索了 `{self.song_as_jsonline(song)}`\n'
                        '应用返回的候选项如下：\n'
                        f'{self.standby_list_as_jsonlines(standby_list)}')
        }
        logger.info(f"Try to find {song} standby, user msg: {user_msg['content']}")
        client = self.ai.get_async_client()
        try:
            stream = await client.chat.completions.create(
                model=self.ai.model,
                messages=[sys_msg, user_msg],
                stream=True,
            )
        except:  # noqa
            logger.exception('AI request failed')
            return []

        rr, rw, wtask = await a_handle_stream(stream)
        source_media_pair = None
        while True:
            try:
                lineb = await rr.readline()
            except:  # noqa
                logger.exception('readline failed')
                break
            if not lineb:
                break

            source_media_pair = await self.try_get_source_media_pair()
            if source_media_pair:
                logger.info(f'Standby from ${source_media_pair[0]} is available')
                await stream.close()
                break

            line = lineb.decode('utf-8')
            logger.debug(f"{song} standby: {line}")
            try:
                song_json = json.loads(line)
            except json.JSONDecodeError:
                logger.debug(f'invalid json line: {line}')
                continue
            if song_json['score'] < self.min_score:
                continue
            try:
                source, identifier = song_json['song_id'].split('___')
            except ValueError:
                logger.error(f'AI returns a invalid response: {song_json}')
                break
            if source in self.song_by_source:  # Only use the first song for each source.
                continue
            for standby in standby_list:
                if standby.source == source and standby.identifier == identifier:
                    self.song_by_source[source] = standby
                    task = run_afn(
                        self._prepare_media, standby, self.audio_select_policy)
                    self.fetch_media_tasks[source] = task

        try:
            await wtask
        except:  # noqa
            if not source_media_pair:
                logger.exception('Stream consumer error')
        rw.close()
        await rw.wait_closed()

        if not source_media_pair:
            source_media_pair = await self.wait_and_get_source_media_pair()
        if source_media_pair:
            self.cancel_fetch_media_tasks()
            standby = self.song_by_source[source_media_pair[0]]
            media = source_media_pair[1]
            logger.debug(f'Standby matched for {song}')
            return [(standby, media)]
        logger.debug(f'No standby matched for {song}')
        return []

    async def try_get_source_media_pair(self):
        source_media_pair_ = None
        for source, task in self.fetch_media_tasks.items():
            if task.done():
                _media = task.result()
                if _media is not None:
                    source_media_pair_ = (source, _media)
                    break
        return source_media_pair_

    async def wait_and_get_source_media_pair(self):
        fs = self.fetch_media_tasks.values()
        # TODO: add timeout
        for future in as_completed(fs, timeout=None):
            try:
                media = await future
            except:  # noqa
                logger.exception('fetch media task failed')
                continue
            else:
                # When a provider does not implement search method, it returns None.
                source = None
                for source, task in self.fetch_media_tasks.items():
                    if future == task:
                        source = source
                        break
                if media is not None:
                    return (source, media)
                else:
                    logger.debug('No media available in source:${source}')
        return None

    def cancel_fetch_media_tasks(self):
        for source, task in self.fetch_media_tasks.items():
            if not task.done():
                task.cancel()

    @classmethod
    def song_as_jsonline(cls, song: 'BriefSongModel'):
        js = {
            'title': song.title,
            'artists_name': song.artists_name,
        }
        if song.album_name:
            js['album_name'] = song.album_name,
        if song.duration_ms:
            js['duration_ms'] = song.duration_ms,
        return json.dumps(js)

    @classmethod
    def standby_list_as_jsonlines(cls, songs: List['BriefSongModel']):
        lines = []
        for song in songs:
            js = {
                'song_id': f'{song.source}___{song.identifier}',
                'title': song.title,
                'artists_name': song.artists_name,
                'album_name': song.album_name,
                'duration_ms': song.duration_ms,
            }
            lines.append(json.dumps(js))
        return '\n'.join(lines)


# For debugging.
if __name__ == '__main__':
    import os

    from feeluown.library import Library
    from feeluown.library.uri import resolve, Resolver
    from fuo_ytmusic.provider import provider as p1
    from fuo_qqmusic import provider as p2
    from fuo_netease import provider as p3

    logging.basicConfig(level=logging.DEBUG)

    library = Library()
    p1.setup_http_proxy('http://127.0.0.1:7890')
    library.register(p1)
    library.register(p2)
    library.register(p3)
    Resolver.library = library
    library.setup_ai(AI(
        # base_url='https://api.deepseek.com',
        # api_key=os.environ.get('DEEPSEEK_API_KEY'),
        # model='deepseek-chat',
        base_url='https://ark.cn-beijing.volces.com/api/v3',
        api_key=os.environ.get('ARK_API_KEY'),
        model='ep-20250202091715-vwjw2',
        # base_url='https://open.bigmodel.cn/api/paas/v4/',
        # api_key=os.environ.get('GLM_API_KEY'),
        # model='GLM-4-Air',
        # api_key=os.environ.get('MOONSHOT_API_KEY'),
        # base_url='https://api.moonshot.cn/v1',
        # model='moonshot-v1-8k',
    ))
    standby_text = '''\
fuo://qqmusic/songs/409175284   # "下雨天 - 鱼天邻制作" - 南拳妈妈 - "" - 00:51
fuo://qqmusic/songs/463631892   # 下雨天 (DJ阿智版) - 南拳妈妈 & DJ阿智 - "" - 01:28
fuo://qqmusic/songs/102697748   # 下雨天 - 南拳妈妈 - 优の良曲 南搞小孩 - 04:13
fuo://ytmusic/songs/XkcKycMblaE # 下雨天 - 芝麻Mochi - 下雨天 - 04:26
fuo://ytmusic/songs/F-YMyH74748 # 下雨天 - "" - 下雨天 - 04:14
fuo://ytmusic/songs/BBe-Zwb7ElM # Rainy Day (下雨天) - Nan Quan Mama - 優的良曲南搞小孩 - 04:14
fuo://netease/songs/1382202727  # 下雨天（翻自 南拳妈妈NQMM） - 33没事儿 - 翻唱 - 04:10
fuo://netease/songs/2135170282  # 下雨天（Alqas 版） - Kk - 南拳妈妈-下雨天 - 03:26
fuo://netease/songs/1905457762  # 下雨天（Cover 南拳妈妈） - 张贤静 - For you - 04:36
'''
    standby_list = []
    for line in standby_text.splitlines():
        l_nospace = line.strip()
        if l_nospace:
            standby_list.append(resolve(l_nospace))

    print(standby_text)

    matcher = AIStandbyMatcher(
        library.ai, library.a_song_prepare_media_no_exc, 60, '>>>')

    song = BriefSongModel(
        source='dummy',
        identifier='xxx',
        title='下雨天',
        artists_name='南拳妈妈'
    )
    pair = asyncio.run(matcher.match(song, standby_list))
    print(pair)



from feeluown.library import ModelType
from feeluown.utils.utils import DedupList


class RecentlyPlayed:
    """
    RecentlyPlayed records recently played models, currently including songs
    and videos. Maybe artists and albums will be recorded in the future.
    """
    def __init__(self, playlist):
        # Store recently played songs. Newest song is appended to left.
        self._songs = DedupList()

        playlist.song_changed_v2.connect(self._on_song_played)

    def init_from_models(self, models):
        for model in models:
            if ModelType(model.meta.model_type) is ModelType.song:
                self._songs.append(model)

    def list_songs(self):
        """List recently played songs (list of BriefSongModel).
        """
        return list(self._songs.copy())

    def _on_song_played(self, song, _):
        # Remove the song and place the song at the very first if it already exists.
        if song is None:
            return
        if song in self._songs:
            self._songs.remove(song)
        if len(self._songs) >= 100:
            self._songs.pop()
        self._songs.insert(0, song)



from typing import TYPE_CHECKING
from PyQt5.QtWidgets import QWidget, QLabel, QHBoxLayout

from feeluown.gui.widgets.labels import ProgressLabel, DurationLabel
from feeluown.gui.widgets.progress_slider import ProgressSlider

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class PlayerProgressRatioLabel(QWidget):
    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        self._app = app

        self.duration_label = DurationLabel(app, parent=self)
        self.progress_label = ProgressLabel(app, parent=self)

        font = self.font()
        font.setPixelSize(10)
        for label in (self.duration_label, self.progress_label):
            label.setFont(font)
            label.setMinimumWidth(26)

        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(2)
        self._layout.addWidget(self.progress_label)
        self._layout.addWidget(QLabel('/'))
        self._layout.addWidget(self.duration_label)


class PlayerProgressSliderAndLabel(QWidget):
    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)

        self._app = app
        self.slider = ProgressSlider(app, parent=self)
        self.label = PlayerProgressRatioLabel(app, parent=self)

        self._layout = QHBoxLayout(self)
        self._layout.addWidget(self.slider)
        self._layout.addWidget(self.label)



from feeluown.utils.reader import wrap
from feeluown.gui.page_containers.table import Renderer


class LibraryRenderer(Renderer):
    def __init__(self, songs, albums, artists):
        self.songs = songs
        self.albums = albums
        self.artists = artists

    async def render(self):
        self.meta_widget.show()
        self.tabbar.show()
        self.tabbar.library_mode()

        # render songs
        self.show_songs(reader=wrap(self.songs), show_count=True)

        # bind signals
        self.toolbar.filter_albums_needed.connect(
            lambda types: self.albums_table.model().filter_by_types(types))
        self.tabbar.show_songs_needed.connect(
            lambda: self.show_songs(reader=wrap(self.songs),
                                    show_count=True))
        self.tabbar.show_albums_needed.connect(lambda: self.show_albums(self.albums))
        self.tabbar.show_artists_needed.connect(lambda: self.show_artists(self.artists))
